

class Dispatchable:

    def __init__(self):
        self.__dispatcher = None

    @property
    def dispatcher(self):
        return self.__dispatcher

    @dispatcher.setter
    def dispatcher(self, dispatcher):
        self.__dispatcher = dispatcher


from .base_element import ElementConfig

class Tab:

    def __init__(self, automator, label_elem, content_relation_attr, content_relation_type):
        super().__init__(automator, parent)
        self.__label_elem = label_elem
        self.__tab_group = tab_group
        self.__label = label_elem.text

        # Find an element based on content of attribute that ties a tab label to content that it controls.
        attr_value = label_elem.get_attr_value(content_relation_attr)
        with_obj = getattr(With, content_relation_type.name.lower())(attr_value)

        self.__tab_content_lmd = self.automator.create_lmd(with_obj)
        self.__tab_content = None

    @property
    def label(self):
        return self.__name

    def click(self):
        self.__label_elem.click()
        self.__tab_content = self.__tab_grouptab_group.root.element(self.__tab_content_lmd)

    def element(self, *with_locators):
        self.__content.element(*with_locators)


class TabGroup(ElementConfig):

    def __init__(self, automator, emd, *, tab_header_lmd, content_relation_attr, content_relation_type, parent=None):
        super().__init__(automator)
        self.__root = automator.define_element(emd)
        tabs = self.root.define_element(tab_header_lmd).identify()
        self.__tabs = []
        for i in range(tabs.length):
            self.__tabs.append(Tab(self, tabs.at_index(i), content_relation_attr, content_relation_type))

    @property
    def root(self):
        return self.__root

    @property
    def labels(self):
        return [t.label for t in self.__tabs]



import random
import os

from arjuna.interact.gui.auto.base.locatable import Locatable
from arjuna.interact.gui.auto.base.dispatchable import Dispatchable
from arjuna.interact.gui.auto.base.configurable import Configurable
from arjuna.interact.gui.auto.element.guielement import GuiElement
from arjuna.interact.gui.auto.source.parser import *

class _GuiPartialElement(GuiElement):

    def __init__(self, gui, multi_element, index: int, dispatcher_element, iconfig=None):
        super().__init__(gui, multi_element.lmd, iconfig=iconfig)
        self.__multi_element = multi_element
        self.__index = index
        self.dispatcher = dispatcher_element

    def create_dispatcher(self):
        pass

    def _is_partial_element(self):
        return True

    @property
    def index(self):
        return self.__index

class GuiMultiElement(Locatable,Dispatchable,Configurable):
    
    def __init__(self, gui, lmd, iconfig=None): #, parent=None):
        Locatable.__init__(self, gui, lmd) #, parent)
        Dispatchable.__init__(self)
        Configurable.__init__(self, gui, iconfig)
        self.__instance_count = 0
        self.__instances = None
        self.__source_parser = None

        

    # def configure_partial_elements(self, elem_config):
    #     '''
    #         This method is supposed to be called when multielement identification is completed.
    #         This is not used for usual multi-element.
    #         It is used by RadioGroup or DropDown etc which are higher level abstractions that use ME.
    #     '''
    #     for instance in self.__instances:
    #         instance.configure(elem_config)

    # def find(self):
    #     self.parent_container.find_multielement(self)
    #     self.__load_source_parser()

    # #Override
    # def find_if_not_found(self):
    #     if not self.is_found():
    #         self.find()

    def load_source_parser(self):
        self.__source_parser = MultiElementSource()
        for instance in self.__instances:
            instance.load_source_parser()
        self.__source_parser.load(self.__instances)

    def __getitem__(self, index):
        # self.find_if_not_found()
        return self.__instances[index]

    @property
    def instance_count(self):
        # self.find_if_not_found()
        return self.__instance_count

    @instance_count.setter
    def instance_count(self, count):
        self.__instance_count = count
        self.__instances = [_GuiPartialElement(self.gui, self, i, self.dispatcher.get_element_at_index(i), iconfig=self.settings) for i in range(self.instance_count)]

    length = instance_count

    @property
    def random_element(self):
        # self.find_if_not_found()
        return self[random.randint(0, self.instance_count-1)]

    @property
    def first_element(self):
        return self[0]

    @property
    def last_element(self):
        return self[self.length-1]

    def get_element_at_ordinal(self, ordinal):
        # self.find_if_not_found()
        return self.__instances[ordinal-1]

    def get_instance_by_visible_text(self, text):
        texts = self.__get_all_texts()
        first_index = self.__find_first_text_index(texts, text)
        return self[first_index]

    def get_instance_by_value(self, value):
        values = self.__get_all_values()
        first_index = self.__find_first_value_index(values, value)
        return self[first_index]

    def wait_until_visible(self):
        self.find_if_not_found()
        # Should the logic be keptin Setu itself for this????

    def __return_attr_values(self, response):
        if "data" in response and "attrValues" in response["data"]:
            return response["data"]["attrValues"]
        else:
            return None

    def are_selected(self):
        return [instance.is_selected() for instance in self.__instances]

    # getting index attribute when it does not exist retursn value attribute.
    # So, not going the Selenium way. Setu would treat index as computer counting.
    def has_index_selected(self, index):
        # self.find_if_not_found()
        return self[index].is_selected()

    # Ordinal is human counting
    def has_ordinal_selected(self, ordinal):
        return self.has_index_selected(ordinal-1)

    def __find_first_match_index(self, in_sequence, to_match):
        try:
            return in_sequence.index(to_match)
        except:
            return -1

    def __get_all_texts(self):
        # self.find_if_not_found()
        self.load_source_parser()
        texts = self.__source_parser.get_text_contents()
        return texts

    def __find_first_text_index(self, texts, text):
        first_index = texts.index(text)
        if first_index == -1:
            raise Exception("No option with {} visible text present in drop down.".format(text))
        return first_index

    def __get_all_values(self):
        # self.find_if_not_found()
        self.load_source_parser()
        values = self.__source_parser.get_values()
        return values

    def __find_first_value_index(self, values, value):
        first_index = values.index(value)
        if first_index == -1:
            raise Exception("No option with {} value present in drop down.".format(value))
        return first_index

    def get_first_selected_instance(self):
        # self.find_if_not_found()
        self.load_source_parser()
        booleans = self.are_selected()
        first_index = None
        try:
            first_index = booleans.index(True)
        except:
            raise Exception("No option in drop down is currenlty selected.")
        else:
            return self[first_index]

    @property
    def source(self):
        return self.__source_parser



class WebAlertHandler:

    def __init__(self, automator):
        self.__automator = automator
        self.__alert_present = False
        self.__alert = None

    @property
    def automator(self):
        return self.__automator

    def create_alert(self):
        if self.__alert_present:
            return self.__alert
        else:
            self.wait()
            from arjuna.interact.gui.auto.component.webalert import WebAlert
            self.__alert_present = True
            alert = WebAlert(self.automator)
            self.__alert = alert
            return alert

    def delete_alert(self):
        self.__alert_present = False
        self.__alert = None

    def wait(self):
        self.automator.conditions.AlertIsPresent().wait(max_wait_time=self.automator.config.get_arjuna_option_value("guiauto.max.wait").as_int())

    def is_alert_present(self):
        return self.automator.dispatcher.is_web_alert_present()







class Locatable:

    def __init__(self, gui, emd):
        # super().__init__(automator.config) #, obj_name)
        self.__gui = gui
        self.__automator = gui.automator
        self.__lmd = emd
        self.__located = False
        self.__located_by = None 

        # self.__parent = parent

    @property
    def gui(self):
        return self.__gui
        
    @property
    def automator(self):
        return self.__automator

    @property
    def lmd(self):
        return self.__lmd

    @property
    def located(self):
        return self.__located

    @property
    def located_with(self):
        return self.__located_by

    @located_with.setter
    def located_with(self, locator_tuple):
        self.__located = True
        self.__located_by = locator_tuple


'''
From Base Element
    def __init__(self, automator, emd):
        self.__dispatcher_element = None



    # @property
    # def parent_container(self):
    #     if self.__parent is not None:
    #         return self.__parent
    #     else:
    #         return self.__automator

    def _create_element_flat_or_nested(self, locator_meta_data):
        from arjuna.interact.gui.auto.element.guielement import GuiElement
        return GuiElement(self.__automator, locator_meta_data, parent=self) 

    def _create_multielement_flat_or_nested(self, locator_meta_data):
        from arjuna.interact.gui.auto.element.multielement import GuiMultiElement
        return GuiMultiElement(self.__automator, locator_meta_data, parent=self) 

    def create_dispatcher(self):
        self._set_dispatcher(self.dispatcher_creator.create_gui_element_dispatcher(self.__automator.dispatcher, self.setu_id))

'''




class NestedGuiElement:
    '''
        It is not an element container. It means it can not be find elements within itself.
    '''

    def __init__(self, automator, parent_element, emd):
        pass

    



import time
import traceback

from arjuna.core.exceptions import *

# This code is inspired by fluent wait concept in Selenium Webdriver
# Reference code: https://github.com/browserstack/selenium-webdriver-python/edit/master/selenium/webdriver/support/wait.py
# However, rather than a Selnium WebDriver specific and driver level concept,
# Setu aims to make it a more generic concept available beyond WebDriver and GUI automation.
# Setu makes waiting accessible through a condition rather than the other way round.
# i.e. object.condition(args).wait(), rather than wait(driver).condition(args)

class ConditionException(Exception):

    def __init__(self, msg):
        super().__init__(msg)

class Condition:

    def __init__(self, dynamic_caller, *args, **kwargs):
        self.__dynamic_caller = dynamic_caller
        self.__args = args
        self.__kwargs = kwargs
        self.__call_result = None

    def get_call_result(self):
        return self.__call_result

    def wait(self, *, max_wait_time=60, poll_interval=0.5):
        end_time = time.time() + max_wait_time
        e = None
        etrace = None
        while(True):
            try:
                if self.is_met():
                    return self.get_call_result()
            except ConditionException as ce:
                raise ce
            except WaitableError as we:
                e = we
                import traceback
                etrace = traceback.format_exc()
                pass
            except Exception as f:
                import traceback
                traceback.print_exc()
                raise Exception("An unexpected exception occured in dynamic wait. Message: {}".format(str(f)))
            time.sleep(poll_interval)
            ctime = time.time()
            if(ctime > end_time):
                break
        raise TimeoutError(self.__class__.__name__, str(e) + etrace)

    def execute(self):
        self.__call_result = self.__dynamic_caller.call(*self.__args, **self.__kwargs)

class CommandCondition(Condition):

    def __init__(self, dynamic_caller, returns_none=True):
        super().__init__(dynamic_caller)
        self.__returns_none = returns_none

    def is_met(self):
        self.execute()
        if self.__returns_none: return True
        return self.get_call_result() is not None

class BooleanCondition(Condition):

    def __init__(self, dynamic_caller, expectation_type=True):
        super().__init__(dynamic_caller)
        self.__expectation_type = expectation_type

    def is_met(self):
        try:
            self.execute()
            return self.__expectation_type == self.get_call_result()
        except Exception as e:
            import traceback
            traceback.print_exc()
            raise ConditionException("Unexpected exception in boolean condition checking: " + str(e))


from .enums import TimeUnit

class Time:

    def __init__(self, time_unit, value):
        self.__unit = time_unit
        self.__value = value

    @staticmethod
    def seconds(self, secs):
        return Time(TimeUnit.SECONDS, secs)

    @staticmethod
    def milli_seconds(self, ms):
        return Time(TimeUnit.MILLI_SECONDS, ms)

    @staticmethod
    def minutes(self, mins):
        return Time(TimeUnit.MINUTES, mins)

logger = Arjuna.get_logger()
console = Arjuna.get_console()




from enum import Enum, auto


class ArjunaOption(Enum):
    ARJUNA_ROOT_DIR = auto()
    ARJUNA_EXTERNAL_TOOLS_DIR = auto()
    ARJUNA_EXTERNAL_IMPORTS_DIR = auto()
    PYTHON_LOG_NAME = auto()
    LOG_NAME = auto()

    LOG_DIR = auto()
    LOG_FILE_LEVEL = auto()
    LOG_CONSOLE_LEVEL = auto()

    PROJECT_NAME = auto()
    PROJECT_ROOT_DIR = auto()
    PROJECT_CONF_FILE = auto()
    PROJECT_TESTS_DIR = auto()

    DATA_DIR = auto()
    DATA_SOURCES_DIR = auto()
    DATA_REFERENCES_DIR = auto()
    SCREENSHOTS_DIR = auto()
    CONFIG_DIR = auto()

    SETU_PROJECT_DIRS_FILES = auto()
    REPORT_DIR = auto()
    ARCHIVES_DIR = auto()

    AUT_BASE_URL = auto()

    TESTRUN_ENVIRONMENT = auto()
    TESTRUN_HOST_OS = auto()

    SETU_GUIAUTO_ACTOR_MODE = auto()
    SETU_GUIAUTO_ACTOR_URL = auto()

    AUTOMATOR_NAME = auto()

    BROWSER_NAME = auto()
    BROWSER_VERSION = auto()
    BROWSER_MAXIMIZE = auto()
    BROWSER_DIM_HEIGHT = auto()
    BROWSER_DIM_WIDTH = auto()
    BROWSER_BIN_PATH = auto()
    BROWSER_PROXY_ON = auto()

    GUIAUTO_INPUT_DIR = auto()
    GUIAUTO_NAMESPACE_DIR = auto()
    GUIAUTO_DEF_MULTICONTEXT = auto()
    GUIAUTO_CONTEXT = auto()
    SCROLL_PIXELS = auto()
    SWIPE_TOP = auto()
    SWIPE_BOTTOM = auto()
    SWIPE_MAX_WAIT = auto()
    GUIAUTO_MAX_WAIT = auto()
    GUIAUTO_SLOMO_ON = auto()
    GUIAUTO_SLOMO_INTERVAL = auto()

    MOBILE_OS_NAME = auto()
    MOBILE_OS_VERSION = auto()
    MOBILE_DEVICE_NAME = auto()
    MOBILE_DEVICE_UDID = auto()
    MOBILE_APP_FILE_PATH = auto()

    SELENIUM_DRIVER_PROP = auto()
    SELENIUM_DRIVERS_DIR = auto()
    SELENIUM_DRIVER_PATH = auto()

    APPIUM_HUB_URL = auto()
    APPIUM_AUTO_LAUNCH = auto()

    IMAGE_COMPARISON_MIN_SCORE = auto()

    UNITEE_PROJECT_DIRS_FILES = auto()
    UNITEE_PROJECT_SESSIONS_DIR = auto()
    UNITEE_PROJECT_GROUPS_DIR = auto()
    UNITEE_PROJECT_TESTS_DIR = auto()
    UNITEE_PROJECT_TEST_MODULE_IMPORT_PREFIX = auto()
    UNITEE_PROJECT_FIXTURES_IMPORT_PREFIX = auto()
    UNITEE_PROJECT_CORE_DIR = auto()
    UNITEE_PROJECT_CORE_DB_CENTRAL_DIR = auto()
    UNITEE_PROJECT_CORE_DB_CENTRAL_DBFILE = auto()
    UNITEE_PROJECT_CORE_DB_RUN_DBFILE = auto()
    UNITEE_PROJECT_CORE_DB_ALLRUN_DIR = auto()
    UNITEE_PROJECT_CORE_DB_TEMPLATE_DIR = auto()
    UNITEE_PROJECT_CORE_DB_TEMPLATE_CENTRAL_DBFILE = auto()
    UNITEE_PROJECT_CORE_DB_TEMPLATE_RUN_DBFILE = auto()
    UNITEE_PROJECT_REPORTER_MODE = auto()
    UNITEE_PROJECT_ACTIVE_REPORTERS = auto()
    UNITEE_PROJECT_DEFERRED_REPORTERS = auto()
    UNITEE_PROJECT_FAILFAST = auto()
    UNITEE_PROJECT_REPORT_NAME_FORMAT = auto()
    UNITEE_PROJECT_REPORT_HEADERS_TMETA = auto()
    UNITEE_PROJECT_REPORT_HEADERS_IGMETA = auto()
    UNITEE_PROJECT_REPORT_HEADERS_PROPS = auto()
    UNITEE_PROJECT_REPORT_HEADERS_REPORTABLE_TEST = auto()
    UNITEE_PROJECT_REPORT_HEADERS_REPORTABLE_STEP = auto()
    UNITEE_PROJECT_REPORT_HEADERS_REPORTABLE_ISSUE = auto()
    UNITEE_PROJECT_REPORT_HEADERS_REPORTABLE_IGNORED = auto()
    UNITEE_PROJECT_REPORT_HEADERS_REPORTABLE_FIXTURE = auto()
    UNITEE_PROJECT_REPORT_HEADERS_REPORTABLE_EVENT = auto()
    UNITEE_PROJECT_RUNID = auto()
    UNITEE_PROJECT_IRUNID = auto()
    UNITEE_PROJECT_SESSION_NAME = auto()
    UNITEE_PROJECT_CORE = auto()
    UNITEE_PROJECT_SCREENSHOTS_RUN_DIR = auto()
    UNITEE_PROJECT_RUN_REPORT_DIR = auto()
    UNITEE_PROJECT_RUN_REPORT_JDB_DIR = auto()
    UNITEE_PROJECT_RUN_REPORT_JSON_DIR = auto()
    UNITEE_PROJECT_RUN_REPORT_JSON_TESTS_DIR = auto()
    UNITEE_PROJECT_RUN_REPORT_JSON_IGNOREDTESTS_DIR = auto()
    UNITEE_PROJECT_RUN_REPORT_JSON_ISSUES_DIR = auto()
    UNITEE_PROJECT_RUN_REPORT_JSON_EVENTS_DIR = auto()
    UNITEE_PROJECT_RUN_REPORT_JSON_FIXTURES_DIR = auto()


class Key(Enum):
	NULL = auto()
	CANCEL = auto()
	HELP = auto()
	BACKSPACE = auto()
	BACK_SPACE = auto()
	TAB = auto()
	CLEAR = auto()
	RETURN = auto()
	ENTER = auto()
	SHIFT = auto()
	LEFT_SHIFT = auto()
	CONTROL = auto()
	LEFT_CONTROL = auto()
	ALT = auto()
	LEFT_ALT = auto()
	PAUSE = auto()
	ESCAPE = auto()
	SPACE = auto()
	PAGE_UP = auto()
	PAGE_DOWN = auto()
	END = auto()
	HOME = auto()
	LEFT = auto()
	ARROW_LEFT = auto()
	UP = auto()
	ARROW_UP = auto()
	RIGHT = auto()
	ARROW_RIGHT = auto()
	DOWN = auto()
	ARROW_DOWN = auto()
	INSERT = auto()
	DELETE = auto()
	SEMICOLON = auto()
	EQUALS = auto()

	NUMPAD0 = auto()
	NUMPAD1 = auto()
	NUMPAD2 = auto()
	NUMPAD3 = auto()
	NUMPAD4 = auto()
	NUMPAD5 = auto()
	NUMPAD6 = auto()
	NUMPAD7 = auto()
	NUMPAD8 = auto()
	NUMPAD9 = auto()
	MULTIPLY = auto()
	ADD = auto()
	SEPARATOR = auto()
	SUBTRACT = auto()
	DECIMAL = auto()
	DIVIDE = auto()

	F1 = auto()
	F2 = auto()
	F3 = auto()
	F4 = auto()
	F5 = auto()
	F6 = auto()
	F7 = auto()
	F8 = auto()
	F9 = auto()
	F10 = auto()
	F11 = auto()
	F12 = auto()

	META = auto()
	COMMAND = auto()

class ModifierKey(Enum):
    CTRL = auto()
    CMD = auto()
    XCTRL = auto()
    ALT = auto()
    SHIFT = auto()

class TimeUnit(Enum):
    MILLI_SECONDS = auto()
    SECONDS = auto()
    MINUTES = auto()

class DesktopOS(Enum):
    WINDOWS = auto()
    MAC = auto()

class SetuActorMode(Enum):
    LOCAL = auto()
    REMOTE = auto()

class MobileOsName(Enum):
    ANDROID = auto()
    IOS = auto()

class BrowserName(Enum):
    CHROME = auto()
    FIREFOX = auto()
    SAFARI = auto()
    IE = auto()
    OPERA = auto()
    HTML = auto()

class AppiumAndroidBrowserName(Enum):
    BROWSER = auto()
    CHROME = auto()

class AppiumIosBrowserName(Enum):
    SAFARI = auto()

class FileFormat(Enum):
    INI = auto()
    TXT = auto()
    DELIMITED = auto()
    XLS = auto()
    CSV = auto()
    GNS = auto()

class DataFileFormat(Enum):
    INI = auto()
    TXT = auto()
    DELIMITED = auto()
    XLS = auto()
    CSV = auto()

class Order(Enum):
    RETAIN = auto()
    BY_NAME = auto()
    RANDOM = auto()

class Filter(Enum):
    INCLUDE = auto()
    EXCLUDE = auto()

class GuiAutomationContext(Enum):
    WEB = auto()
    NATIVE = auto()
    SCREEN = auto()
    ANDROID_WEB = auto()
    IOS_WEB = auto()
    ANDROID_NATIVE = auto()
    IOS_NATIVE = auto()

    # DESKTOP_CONTEXTS = {GuiAutomationContext.NATIVE, GuiAutomationContext.WEB}
    # MOBILE_WEB_CONTEXTS = {GuiAutomationContext.ANDROID_WEB, GuiAutomationContext.IOS_WEB}
    # ALL_WEB_CONTEXTS = {GuiAutomationContext.WEB, GuiAutomationContext.ANDROID_WEB, GuiAutomationContext.IOS_WEB}
    # MOBILE_NATIVE_CONTEXTS = {GuiAutomationContext.ANDROID_NATIVE, GuiAutomationContext.IOS_NATIVE}
    
    @staticmethod
    def isDesktopContext(context):
        return context in GuiAutomationContext.DESKTOP_CONTEXTS

    @staticmethod
    def is_mobile_web_context(context):
        return context in GuiAutomationContext.MOBILE_WEB_CONTEXTS

    @staticmethod
    def is_mobile_native_context(context):
        return context in GuiAutomationContext.MOBILE_NATIVE_CONTEXTS

    @staticmethod
    def is_web_context(context):
        return context in GuiAutomationContext.ALL_WEB_CONTEXTS

class GuiAutomatorName(Enum):
    SELENIUM = auto()
    APPIUM = auto()

class GuiElementType(Enum):
    TEXTBOX = auto()
    PASSWORD = auto()
    LINK = auto()
    BUTTON = auto()
    SUBMIT_BUTTON = auto()
    DROPDOWN = auto()
    CHECKBOX = auto()
    RADIO = auto()
    IMAGE = auto()

class OS(Enum):
	WINDOWS = auto()
	MAC = auto()
	LINUX = auto()
	ANDROID = auto()
	IOS = auto()

class MobileView(Enum):
    NATIVE_APP = auto()
    WEBVIEW = auto()

class Device(Enum):
    PC = auto()
    MOBILE = auto()
    GENERIC = auto()    

class Device(Enum):
    UP = auto()
    DOWN = auto()
    LEFT = auto()
    RIGHT = auto()



class LoggingLevel(Enum):
    DEBUG = auto()
    INFO = auto()
    WARN = auto()
    ERROR = auto()
    FATAL = auto()

class ConfigPropertyFormattingTypeEnum(Enum):
    PATH_TO_ABS_PATH = auto()

class DeviceTypeEnum(Enum):
    PC = auto()
    MOBILE = auto()
    GENERIC = auto()

class HoconSyntaxTypeEnum(Enum):
    PROPERTIES = auto()
    JSON = auto()
    CONF = auto()


class FilterTypeEnum(Enum):
    INCLUDE = auto()
    EXCLUDE = auto()


class DiscoveredFileAttributeEnum(Enum):
    NAME = auto()
    EXTENSION = auto()
    FULL_NAME = auto()
    DIRECTORY_ABSOLUTE_PATH = auto()
    DIRECTORY_RELATIVE_PATH = auto()
    PACKAGE_DOT_NOTATION = auto()
    COMMA_SEPATARED_RELATIVE_PATH = auto()
    CONTAINER = auto()
    CONTAINER_TYPE = auto()

class ValueType(Enum):
    BOOLEAN = auto()
    STRING = auto()
    STRING_LIST = auto()
    NONE = auto()
    NUMBER = auto()
    NUMBER_LIST = auto()
    LIST = auto()
    ANYREF = auto()
    ENUM = auto()
    ENUM_LIST = auto()
    INTEGER = auto()
    OBJECT_LIST = auto()
    FLOAT = auto()
    DOUBLE = auto()
    LONG = auto()
    NOT_SET = auto()
    NA = auto()
    INT_LIST = auto()

class NamesContainerTypeEnum(Enum):
    TEST = auto()
    TEST_RESULT = auto()
    IGNORED_TEST = auto()
    STEP_RESULT = auto()
    ISSUE = auto()
    DEFAULT_FIXTURE_tfuncs = auto()
    COMPONENT_NAMES = auto()
    TEST_OBJECT = auto()
    EXCLUDED_TEST_RESULT = auto()
    EVENT = auto()
    TEST_OTYPE_NAMES = auto()
    FIXTURE_RESULT = auto()

class ConfigPropertyLevelEnum(Enum):
    CENTRAL = auto()
    THREAD = auto()

class CorePropertyTypeEnum(Enum):
    ARJUNA_ROOT_DIR = auto()
    PROG = auto()
    CONFIG_CENTRAL_FILE_NAME = auto()
    CONFIG_PROJECTS_DIR = auto()
    WORKSPACE_DIR = auto()
    EXTERNAL_TOOLS_DIR = auto()
    EXTERNAL_IMP_DIR = auto()
    LOGGER_DIR = auto()
    CONFIG_DIR = auto()
    LOGGER_CONSOLE_LEVEL = auto()
    LOGGER_FILE_LEVEL = auto()
    LOGGER_NAME = auto()
    PROJECT_DIRS_FILES = auto()

class GuiInteractionConfigType(Enum):
    CHECK_TYPE = auto()
    CHECK_PRE_STATE = auto()
    CHECK_POST_STATE = auto()


import time
import inspect

class HardCoded:

    @classmethod
    def __get_invoker(cls):
        frame = inspect.stack()[2]
        mod = inspect.getmodule(frame[0])
        mod_name = mod.__name__
        mod_file = mod.__file__
        mod_script = ""
        if mod_name == "__main__":
            mod_script = "Script:<{}> at ".format(mod_file)
        else:
            mod_script = "Module:<{}> File:<{}>".format(mod_name, mod_file)
        func = frame[3]
        if func == "<module>":
            func = ""
        else:
            func = "Function/Method: <{}> in ".format(func)
        line = frame[2]
        return "{}{}Line: {}".format(func, mod_script, line)

    @classmethod
    def __log(cls, invoker, why, seconds):
        from arjuna import Arjuna
        Arjuna.get_logger().warn("Hardcoded sleep executed for {} seconds by {}. Reason by author: {}".format(seconds, invoker, why))

    @classmethod
    def sleep(cls, why, seconds):
        time.sleep(seconds)
        cls.__log(cls.__get_invoker(), why, seconds)


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import functools
import threading
import types

from arjuna.core.adv.wrappers import ClassPassThrough


# From blog post by Daniel Holden at http://theorangeduck.com/page/synchronized-python
def sync_function(func):
    # reentrant lock to supprt recursion
    func.__lock__ = threading.RLock()

    @functools.wraps(func)
    def synced_func(*args, **kws):
        with func.__lock__:
            return func(*args, **kws)

    return synced_func


def sync_class(klass):
    for key in klass.__dict__:
        val = klass.__dict__[key]
        if type(val) is types.FunctionType:
            decorator = sync_function
            setattr(klass, key, decorator(val))

    return klass


# From blog post by Daniel Holden at http://theorangeduck.com/page/synchronized-python
# To use the following you should create self.lock = threading.RLock() in __init__
# Per instance basis lock
def sync_method(lock_name):
    def decorator(method):
        @functools.wraps(method)
        def synced_method(self, *args, **kws):
            lock = getattr(self, lock_name)
            with lock:
                return method(self, *args, **kws)

        return synced_method

    return decorator



from arjuna.interact.gui.auto.base.locatable import Locatable
from arjuna.interact.gui.auto.base.interactable import Interactable
from arjuna.interact.gui.auto.base.container import ElementContainer

class GuiElement(ElementContainer, Locatable, Interactable):

    def __init__(self, gui, emd, iconfig=None):
        ElementContainer.__init__(self, gui.automator.config)
        Locatable.__init__(self, gui, emd) #, parent, obj_name="GuiElement")
        Interactable.__init__(self, gui, iconfig)

    def element(self, *str_or_with_locators, iconfig=None):
        lmd = self.gui.convert_to_with_lmd(*str_or_with_locators)
        return element_with_lmd(self.gui, lmd, iconfig=iconfig)

    def element_with_lmd(self, gui, lmd, iconfig=None):
        from arjuna.interact.gui.auto.element.guielement import GuiElement
        gui_element = GuiElement(self.gui, lmd, iconfig=iconfig)
        self.load_element(gui_element)
        return gui_element        

    def multi_element(self, *str_or_with_locators, iconfig=None):
        lmd = self.gui.convert_to_with_lmd(*str_or_with_locators)
        return self.multi_element_with_lmd(self.gui, lmd, iconfig=iconfig)

    def multi_element_with_lmd(self, gui, lmd, iconfig=None):
        from arjuna.interact.gui.auto.element.multielement import GuiMultiElement
        m_guielement = GuiMultiElement(self.gui, lmd, iconfig=iconfig)
        self.load_multielement(m_guielement)
        return m_guielement

    def find_element_with_js(self, js):
        raise Exception("With.JS is currently not supported for nested element finding.")

    def find_multielement_with_js(self, js):
        raise Exception("With.JS is currently not supported for nested element finding.")



class GuiNameStore:

    def __init__(self):
        # dict<String, GuiNameStore>
        self.__ns_map = {}

    # Needs to be thread safe
    def has_namespace(self, name):
        return name in self.__ns_map

    # loader is GuiNamespaceLoader
    # Needs to be thread safe
    def load_namespace(self, name, loader):
        if not self.has_namespace(name):
            loader.load()
            self.__ns_map[name.lower()] = loader.namespace

        return self.__ns_map[name.lower()]

    # Needs to be thread-safe
    def get_namespace(self, name):
        return self.__ns_map[name.lower()]


from .gui import *

class Page(AppPortion):

    def __init__(self, *, source_gui, label=None):
        # app = isinstance(source_gui, App) and source_gui or source_gui.app
        super().__init__(automator=source_gui.automator, label=label)

class Widget(AppPortion):

    def __init__(self, page, *, label=None):
        super().__init__(automator=page.automator, label=label)   
        self.__page = page

    @property
    def page(self):
        return self.__page

class App(Gui, metaclass=abc.ABCMeta):

    def __init__(self, *, config=None, ext_config=None, label=None, ns_dir=None):
        super().__init__(config=config, ext_config=ext_config, label=label)
        self.__ui = None
        self.__automator = None
        self.__ns_dir = ns_dir

    @property
    def automator(self):
        return self.__automator

    @property
    def ns_dir(self):
        return self.__ns_dir

    def _launchautomator(self):
        # Default Gui automation engine is Selenium
        from arjuna.interact.gui.auto.automator import GuiAutomator
        self.__automator = GuiAutomator(self, self.config, self.ext_config)    

    @property
    def ui(self):
        return self.__ui

    def _create_default_ui(self):
        self.__ui = Page(source_gui=self, label="{}-Def-UI".format(self.label))

    @abc.abstractmethod
    def launch(self):
        pass

    def prepare_widget(self, widget_object):
        return widget_object

class WebApp(App):

    def __init__(self, *, base_url=None, blank_slate=False, config=None, ext_config=None, label=None, ns_dir=None):
        '''
            Creates and returns GuiAutomator object for provided config.
            If no configuration is provided reference configuration is used.
            You can also provide GuiDriverExtendedConfig for extended configuration for WebDriver family of libs. 
        '''
        super().__init__(config=config, ext_config=ext_config, label=label, ns_dir=ns_dir)
        from arjuna.core.enums import ArjunaOption
        self.__base_url = base_url is not None and base_url or self.config.get_arjuna_option_value(ArjunaOption.AUT_BASE_URL).as_str()

    @property
    def base_url(self):
        return self.__base_url

    def launch(self, blank_slate=False):
        self._launchautomator()
        if not blank_slate:
            self.automator.browser.go_to_url(self.base_url)
        self._create_default_ui()

    def quit(self):
        self.automator.quit()


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import pkg_resources
import time
import os
import sys

def join_paths(*paths):
    return os.path.abspath(os.path.join(*paths))

root_dir = join_paths(os.path.dirname(os.path.realpath(__file__)), "..")
importables_dir = join_paths(root_dir, "third_party")

sys.path.insert(0, importables_dir)
sys.path.insert(0, root_dir)

class ArjunFacade:

    def __init__(self):
        self.__version = pkg_resources.require("arjuna")[0].version

    def launch(self, raw_args):
        from arjuna.core.reader.textfile import TextResourceReader
        from arjuna.interface.cli import ArjunaCLI
        reader = TextResourceReader("header.txt")
        print(reader.read().format(version=self.__version))
        reader.close()

        print("Parsing CLI Options...")
        cli = ArjunaCLI(raw_args)
        # Initialize the Arjuna Core as per CLI options
        print("Intializing Arjuna...")
        cli.init()

        print("Executing Arjuna Command...")
        cli.execute()

from arjuna.engine import Arjuna

from arjuna.interact.gui.helpers import With, WithType
from arjuna.interact.gui.helpers import GuiInteractionConfig, GuiDriverExtendedConfigBuilder
from arjuna.interact.gui.helpers import Screen

from arjuna.interact.gui.gom import WebApp, Page, Widget

from arjuna.core.exceptions import *
from arjuna.core.enums import *

from arjuna.core.audit import HardCoded


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import abc
import os

from pyhocon import ConfigFactory
from pyhocon import ConfigTree, config_tree

def escape_value(value):
    return value.replace('=', '\\=').replace('!', '\\!').replace('#', '\\#').replace('\n', '\\\n')


class HoconReader(metaclass=abc.ABCMeta):
    def __init__(self):
        self.__map = {}
        self.__conf = None

    def get_flat_map(self):
        return self.__map

    def get_map(self):
        return self.__conf

    @abc.abstractmethod
    def _load_config(self):
        pass

    def set_config(self, conf):
        self.__conf = conf
    #
    # def __get_primitive(self, config):
    #     return config
    #     # if isinstance(config, config_tree.NoneValue):
    #     #     return NONE
    #     # elif isinstance(config, list):
    #     #     return ObjectListValue(config)
    #     # elif isinstance(config, str):
    #     #     return StringValue(config)
    #     # elif isinstance(config, bool):
    #     #     return BooleanValue(config)
    #     # else:
    #     #     return NumberValue(config)

    # def process(self):
    #     def inner(config, od, key_stack=[], out=None):
    #         if isinstance(config, ConfigTree):
    #             for key, item in config.items():
    #                 if isinstance(item, ConfigTree):
    #                     od[key] = {}
    #                     d = od[key]
    #                 else:
    #                     d = od
    #                 inner(item, d, key_stack + [key], out)
    #         elif isinstance(config, list):
    #             k = key_stack[-1]
    #             out['.'.join(key_stack)] = []
    #             od[k] = []
    #             for index, item in enumerate(config):
    #                 to_append = None
    #                 if item is not None:
    #                     if isinstance(item, ConfigTree):
    #                         d = {}
    #                         inner(item, d, [], {})
    #                         to_append = d
    #                     else:
    #                         to_append = self.__get_primitive(item)
    #                 out['.'.join(key_stack)].append(to_append)
    #                 od[k].append(to_append)
    #         else:
    #             k = key_stack[-1]
    #             p = self.__get_primitive(config)
    #             od[k] = p
    #             # out is None only for the first time call
    #             # if out is empty, it needs to be populated, so DO NOT DO PY BOOL i.e. if out:
    #             if out is not None:
    #                 out['.'.join(key_stack)] = p
    #
    #     self._load_config()
    #     inner(self.__conf, self.__nested, [], self.__map)
    #     # ArjunaCore.console.display(self.__conf)
    #     # ArjunaCore.console.display(self.__map)


    ######################################################################
    # Modified version of to_properties from pychocon package.
    # License details for pyhocon are listed in license.txt file.
    ######################################################################

    def process(self):
        def inner(out, config, key_stack=[]):
            if isinstance(config, ConfigTree):
                for key, item in config.items():
                    if item is not None:
                        inner(out, item, key_stack + [key])
            elif isinstance(config, config_tree.NoneValue):
                out[".".join(key_stack)] = None
            else:
                out[".".join(key_stack)] = config

        self._load_config()
        inner(self.__map, self.__conf)

class HoconConfigDictReader(HoconReader):
    def __init__(self, conf):
        super().__init__()
        self.set_config(conf)

    def _load_config(self):
        pass


class HoconFileReader(HoconReader):
    def __init__(self, path):
        super().__init__()
        self.conf_path = path

    def _load_config(self):
        self.set_config(ConfigFactory.parse_file(self.conf_path))


class HoconResourceReader(HoconFileReader):
    def __init__(self, name):
        # ArjunaCore.console.display(os.path.abspath(os.path.join(os.path.dirname(__file__), os.sep.join(["../..", "res", name]))))
        super().__init__(os.path.abspath(os.path.join(os.path.dirname(__file__), os.sep.join(["../..", "res", name]))))


class HoconStringReader(HoconReader):
    def __init__(self, conf_string):
        super().__init__()
        self.conf_string = conf_string

    def _load_config(self):
        conf = ConfigFactory.parse_string(self.conf_string.replace("\\\\\\\\", "fourslash").replace("\\\\", "twoslash")
                                          .replace("\\", "\\\\")
                                          .replace("twoslash", "\\\\")
                                          .replace("fourslash", "\\\\\\\\"))
        self.set_config(conf)



'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from arjuna.core.exceptions import *

class NamedString:
    def __init__(self, internal, external):
        self.code = internal
        self.name = external

    def get_code(self):
        return self.code

    def get_name(self):
        return self.name


class NamedStringsContainer:
    def __init__(self, name):
        self.container_name = name
        self.names = []

    def add(self, name):
        self.names.append(name)

    def get_name(self):
        return self.container_name

    def get_named_strings(self):
        return self.names


class Name(NamedString):
    pass


class Message(NamedString):
    pass


class NamesContainer(NamedStringsContainer):
    pass


class MessagesContainer(NamedStringsContainer):
    pass


def populate_from_codelist(map, codes):
    for index, code in enumerate(codes):
        map[code[0].upper().trim()] = code[2]


def populate_from_codemap(map, codes):
    for k, v in codes:
        map[k.upper()] = v


def populate_from_namedstringlist(map, named_strings):
    for ns in named_strings:
        map[ns.get_code().upper()] = ns.get_name()


class StringsManager:
    def __init__(self):
        self.msg_map = {}
        self.name_map = {}
        self.prop_map = {}
        self.flattened_names = {}

    def populate_names(self, names_containers_list):
        for nc in names_containers_list:
            if nc.get_name() not in self.name_map:
                self.name_map[nc.get_name()] = {}
            populate_from_namedstringlist(self.name_map[nc.get_name()], nc.get_named_strings())

    def populate_messages(self, messages_containers_list):
        for mc in messages_containers_list:
            if mc.get_name() not in self.msg_map:
                self.msg_map[mc.get_name()] = {}
            populate_from_namedstringlist(self.msg_map[mc.get_name()], mc.get_named_strings())

    def populate_flattened_names(self):
        for section in self.name_map:
            for key in self.name_map[section]:
                self.flattened_names[section + "::" + key] = self.name_map[section][key]

    def get_all_names(self):
        return self.name_map

    def get_all_messages(self):
        return self.msg_map;

    def get_flattneded_names(self):
        return self.flattened_names;

    def __section_exists(self, section):
        return section in self.msg_map

    def __code_exists(self, section, code):
        if not self.__section_exists(section):
            return False;
        return code in self.msg_map[section]

    def __throw_not_initialized_exception(self, context, method):
        raise Problem("adv", context, method, "LOCALIZER_NOT_INITIALIZED",
                      "Strings Manager not initialized.")

    def __get_text_for_code(self, section, msg_code):
        section_code = section.to_upper_case().trim();
        code = msg_code.to_upper_case().trim();
        if not self.__code_exists(section_code, code):
            return code;
        return self.msg_map[section_code][code]

    def get_info_message_text(self, msg_code):
        return self.__get_text_for_code("INFO_MESSAGES", msg_code)

    def get_problem_text(self, msg_code):
        return self.__get_text_for_code("PROBLEM_MESSAGES", msg_code)

    def get_warning_text(self, msg_code):
        return self.__get_text_for_code("WARNING_MESSAGES", msg_code);

    def get_configured_name(self, section_name, internal_name):
        return self.name_map[section_name.to_upper_case().trim()][internal_name.to_upper_case().trim()]

    class problem_codes:
        pass

    class info_codes:
        pass

    class error_codes:
        pass

    def add_property_name(self, prop_code, prop_name):
        self.prop_map[prop_code.upper()] = prop_name

    def get_property_name(self, prop_code):
        return self.prop_map[prop_code.upper()]



import abc

from arjuna.interact.gui.auto.finder.element_finder import ElementFinder

class ElementContainer(metaclass=abc.ABCMeta):

    def __init__(self, config):
        self.__config = config
        from arjuna.interact.gui.auto.condition.container_conditions import GuiElementContainerConditions
        self.__element_finder = ElementFinder(self)
        self.__container_conditions = GuiElementContainerConditions(self)

    @property
    def config(self):
        return self.__config

    @property
    def element_finder(self):
        return self.__element_finder

    @property
    def max_wait_time(self):
        return self.config.get_arjuna_option_value("guiauto.max.wait").as_int()

    def wait_until_element_found(self, gui_element, max_wait_time=None):
        max_wait_time = max_wait_time and max_wait_time or self.max_wait_time
        return self.__container_conditions.PresenceOfElement(gui_element).wait(max_wait_time=max_wait_time)

    def wait_until_multielement_found(self, multi_guielement, max_wait_time=None):
        max_wait_time = max_wait_time and max_wait_time or self.max_wait_time
        return self.__container_conditions.PresenceOfMultiElement(multi_guielement).wait(max_wait_time=max_wait_time)

    def load_multielement(self, multi_guielement, max_wait_time=None):
        locator_type, locator_value, instance_count, dispatcher = self.wait_until_multielement_found(multi_guielement, max_wait_time=max_wait_time)
        if instance_count == 0:
            raise Exception("MultiElement could not be found with any of the provided locators.")
        multi_guielement.located_with = locator_type, locator_value
        multi_guielement.dispatcher = dispatcher
        multi_guielement.instance_count = instance_count
        multi_guielement.load_source_parser()

    def load_element(self, gui_element, max_wait_time=None):
        locator_type, locator_value, instance_count, dispatcher = self.wait_until_element_found(gui_element, max_wait_time=max_wait_time)
        gui_element.located_with = locator_type, locator_value
        gui_element.dispatcher = dispatcher
        gui_element.load_source_parser()


from arjuna.interact.gui.auto.base.configurable import Configurable
from arjuna.interact.gui.auto.base.dispatchable import Dispatchable
from arjuna.interact.gui.auto.source.parser import ElementXMLSourceParser
from arjuna.core.exceptions import GuiElementTextNotSetError

class Interactable(Configurable, Dispatchable):

    def __init__(self, gui, iconfig): #, parent=None, find=True):
        self.__config = gui.automator.config
        Configurable.__init__(self, gui, iconfig)
        Dispatchable.__init__(self)
        self.__source = None

        #self.__automator = automator
        from arjuna.interact.gui.auto.condition.element_conditions import GuiElementConditions, GuiElementLenientInteraction
        self.__conditions_handler = GuiElementConditions(self)
        self.__interaction_handler = GuiElementLenientInteraction(self)
        self.__source_parser = None

        # # For partial element find is False
        # if find:
        #     self.find()

    @property
    def config(self):
        return self.__config

    def __get_attr_value_from_remote(self, attr, optional=False):
        return self.__return_attr_value(self.dispatcher.get_attr_value(attr, optional))

    def __get_source_from_remote(self):
        return self.__get_attr_value_from_remote("outerHTML")

    def load_source_parser(self):
        raw_source = self.__get_source_from_remote()
        self.__source_parser = ElementXMLSourceParser(raw_source)
        self.__source_parser.load()

    # def find(self):
    #     self.parent_container.find_element(self)
    #     self.load_source_parser()

    # def find_without_wait(self):
    #     self.parent_container.find_element_without_wait(self)

    # identify = find
    # reset = find

    def _is_partial_element(self):
        return False

    def _get_instance_number(self):
        raise Exception("Instance number is applicable only for partial gui elements.")

    # def _kwargs(self, **kwargs):
    #     return self.__append_instance_number(kwargs)

    # def _noargs(self):
    #     return self.__append_instance_number({})

    def _only_send_text(self, text):
        self.dispatcher.send_text(text)

    def _only_set_text(self, text):
        self._only_clear_text()
        self._only_send_text(text)
        entered_text = self.dispatcher.get_attr_value("value")
        if entered_text != text:
            raise GuiElementTextNotSetError("Expected text: {}. Actual text is {}".format(text, entered_text))

    def _only_click(self):
        self.dispatcher.click()

    def __return_attr_value(self, result):
        return result and result or None

    def click(self):
        # self.find_if_not_found()
        self.__wait_until_clickable_if_configured()
        self.interactions.Click().wait()
        #self._only_click()

    def __only_hover(self):
        self.dispatcher.hover()

    def hover(self):
        # self.find_if_not_found()
        self.__wait_until_clickable_if_configured()
        self.__only_hover()

    def __conditional_selected_state_click(self, condition_state):
        # self.find_if_not_found()
        selected = self.is_selected()
        if selected == condition_state:
            self.__wait_until_clickable_if_configured()
            self._only_click()

    def select(self):
        self.__conditional_selected_state_click(False)

    def deselect(self):
        self.__conditional_selected_state_click(True)

    # def reload(self):
    #     pass

    # wait_until_present = reload

    def wait_until_visible(self):
        # self.find_if_not_found()
        self.conditions.IsVisible().wait()

    def wait_until_clickable(self):
        # self.find_if_not_found()
        self.conditions.IsClickable().wait()

    def wait_until_selected(self):
        # self.find_if_not_found()
        self.conditions.IsVisible().wait()

    #################################
    ### State Checking
    #################################
    def is_selected(self):
        # self.find_if_not_found()
        return self.dispatcher.is_selected()

    def is_visible(self):
        # self.find_if_not_found()
        return self.dispatcher.is_visible()

    def is_clickable(self):
        # self.find_if_not_found()
        return self.dispatcher.is_clickable()

    @property
    def conditions(self):
        return self.__conditions_handler

    @property
    def interactions(self):
        return self.__interaction_handler

    def __wait_until_clickable_if_configured(self):
        if self._should_check_pre_state(): self.wait_until_clickable()

    # Properties
    @property
    def source(self): # OLD: refind=True, reload=True
        return self.__source_parser

    @property
    def text(self):
        return self.source.content.text

    # Textbox abstraction
    @text.setter
    def text(self, text):
        self.__wait_until_clickable_if_configured()
        self._only_click()
        self._only_clear_text()
        self.interactions.SetText(text).wait()

    def clear_text(self):
        self.__wait_until_clickable_if_configured()
        self._only_click()
        self._only_clear_text()

    def send_text(self, text):
        self.__wait_until_clickable_if_configured()
        self._only_enter_text(text)

    def enter_text(self, text):
        self.__wait_until_clickable_if_configured()
        self._only_click()
        self._only_enter_text(text)

    def _only_clear_text(self):
        self.dispatcher.clear_text()

    def _only_enter_text(self, text):
        self._only_send_text(text)

    def has_entered_text(self, text):
        pass

    def has_entered_partial_text(self, text):
        pass

    # Checkbox abstraction
    def check(self):
        self.select()

    def uncheck(self):
        self.deselect()

    def toggle_checkbox(self):
        self.click()

    def is_checked(self):
        return self.is_selected()

    def get_property_value(self, attr):
        # self.find_if_not_found()
        return self.__get_attr_value_from_remote(attr)

    def perform_action_chain(self, single_action_chain):
        from arjuna.interact.gui.auto.automator.actions import SingleActionChain
        action_chain = SingleActionChain(self.__automator, element=self)
        action_chain.perform(single_action_chain)



import abc
from arjuna.core.exceptions import WaitableError, GuiElementNotFoundError

class ElementFinder:
    def __init__(self, container): #, obj_name=""):
        self.__container = container
        self.__config = container.config
        # self.__obj_name = obj_name

    @property
    def config(self):
        return self.__config

    @property
    def container(self):
        return self.__container

    @abc.abstractmethod
    def _create_element_flat_or_nested(self, lmd):
        pass

    @abc.abstractmethod
    def _create_multielement_flat_or_nested(self, lmd):
        pass

    def find(self, dispatcher_call, gui_element, context="ELEMENT"):
        found = False
        js_call_name = context == "ELEMENT" and "find_element_with_js" or "find_multielement_with_js"
        js_call = getattr(self.container, js_call_name)
        locators = gui_element.lmd.locators
        if context != "ELEMENT":
            if "POINT" in {l.ltype.name for l in locators}:
                raise ConditionException("With.POINT can be used only with GuiElement.")

        we = None
        for locator in locators: 
            try:
                if locator.ltype.name == "POINT":
                    # Assumption here is that this container is automator.
                    instance_count, dispatcher = js_call("return document.elementFromPoint({}, {})".format(*locator.lvalue))
                elif locator.ltype.name == "JAVASCRIPT":
                    instance_count, dispatcher = js_call(locator.lvalue)
                else:
                    instance_count, dispatcher = dispatcher_call(locator.ltype.name, locator.lvalue)
                return locator.ltype.name, locator.lvalue, instance_count, dispatcher
            except WaitableError as e:
                we = e
            except Exception as f:
                raise f
            else:
                we = None
        if not found:
            raise GuiElementNotFoundError("None of the locators worked.", *gui_element.lmd.locators)


import os

from .nsloader import GuiNamespaceLoaderFactory
from arjuna.interact.gui.auto.finder.emd import SimpleGuiElementMetaData, GuiElementMetaData, Locator

class GuiDef:
    '''
        A GuiDef object is attached to a Gui, which in turn is attached to an automator and hence to a fixed auto context.

        Arjuna does a lazy loading of Gui Definitions. This means a context A GuiDef will lead to loading of all contexts centrally, but use only the one it needs to avoid repeat processing.
    '''

    def __init__(self, name_store, automator, label, def_file_path): # namespace_dir, 
        self.__name_store = name_store
        # self.__namespace_dir = namespace_dir
        self.__automator = automator
        self.__config = automator.config
        self.__auto_context = self.config.guiauto_context
        self.__file_def_path = def_file_path
        self.__ns = None
        ns_name = "file_ns::" + self.__file_def_path.lower()
        if name_store.has_namespace(ns_name):
            self.__ns = name_store.get_namespace(ns_name)
        else:
            self.__ns = name_store.load_namespace(
                ns_name, 
                GuiNamespaceLoaderFactory.create_namespace_loader(
                    self.config,
                    self.__file_def_path
            )
        )

        self.__children = []

    @property
    def config(self):
        return self.__config

    # def add_child(self, label, automator, file_def_path):
    #     self.__children.append(
    #         Gui(self.__name_store, self.__namespace_dir, label, self.__automator, file_def_path)
    # )

    def convert_to_lmd(self, *locators):
        final_locators = []
        for raw_locator in locators:
            if raw_locator.wtype.upper().strip() == "GNS_NAME":
                emd = self.__ns.get_meta_data(raw_locator.wvalue, self.__auto_context)
                for loc in emd.raw_locators:
                    if not raw_locator.named_args:
                        final_locators.append(Locator(ltype=loc.ltype, lvalue=loc.lvalue), named_args={})
                    else:
                        final_locators.append(Locator(ltype=loc.ltype, lvalue=loc.lvalue), named_args=raw_locator.named_args)
            else:
                if not raw_locator.named_args:
                    final_locators.append(Locator(ltype=raw_locator.wtype, lvalue=raw_locator.wvalue), named_args={})
                else:
                    final_locators.append(Locator(ltype=raw_locator.ltype, lvalue=raw_locator.lvalue), named_args=raw_locator.named_args)
        import sys
        sys.exit(1)
        return GuiElementMetaData(final_locators)

    def convert_to_with(self, locator):
        from arjuna.interact.gui.helpers import With
        out_list = []
        impl_with = locator.as_impl_locator()
        emd = self.__ns.get_meta_data(impl_with.wvalue, self.__auto_context)
        for loc in emd.raw_locators:
            wobj = getattr(With, loc.ltype.lower()) (loc.lvalue)# e.g. getattr(With, "ID".lower())("abc")
            if locator.named_args:
                wobj.format(**locator.named_args)
            out_list.append(wobj)
        return out_list

    def create_dispatcher(self):
        # Pages don't use any dispatcher
        pass

class GuiFactory:

    @classmethod
    def create_appdef_from_dir(cls, name, automator, app_def_dir):
        from arjuna.core.enums import ArjunaOption
        considered_path = app_def_dir
        if not os.path.isdir(considered_path):
            ns_dir = automator.config.value(ArjunaOption.GUIAUTO_NAMESPACE_DIR)
            full_path = os.path.join(ns_dir, considered_path)
            considered_path = os.path.abspath(full_path)
            if not os.path.isdir(considered_path):
                raise Exception("Provided root definition path is not a directory: {}".format(app_def_dir))

        app = GuiDef(automator, os.path.join(considered_path, "HomePage.gns"), label=name)
        children_dir = os.path.join(considered_path, "children")
        if os.path.isdir(children_dir):
            lfiles = os.listdir(children_dir)
            for f in lfiles:
                cpath = os.path.join(children_dir, f)
                if os.path.isfile(cpath):
                    base_name = os.path.basename(cpath)
                    app.add_child(base_name, cpath)

    @classmethod
    def create_guidef(cls, automator, def_path):
        return GuiDef(automator, def_path)


import copy
import os
import codecs
import sys
import io
import time
# sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding="UTF-8")
# sys.stderr = io.TextIOWrapper(sys.stderr.buffer, encoding="UTF-8")

import logging
from arjuna.core.utils import thread_utils
from arjuna.core.thread.decorators import *
from arjuna.core.utils import sys_utils
from arjuna.core.adv.proxy import ROProxy
from arjuna.core.adv.decorators import singleton
import codecs
import sys


@singleton
class ArjunaSingleton:

    def __init__(self):
        self.__project_root_dir = None
        self.__test_session = None
        self.__ref_config = None
        self.__context_map = dict()
        self.__default_context_name = "default_context"
        self.__run_id = None

        # From central config
        self.prop_enum_to_prop_path_map = {}
        # It is base config object
        self.arjuna_options = {}

        # Check relevance of above and following
        self.prog = "Arjuna"

        self.dl = None
        self.log_file_discovery_info = False
        self._data_references = {}
        self.central_conf = None
        self.__console = None
        self.__logger = None

    @property
    def gui_mgr(self):
        return self.__test_session.gui_manager

    def init(self, project_root_dir, cli_config, run_id):
        from arjuna.configure.impl.processor import ConfigCreator
        ConfigCreator.init()
        self.__project_root_dir = project_root_dir

        from arjuna.session.invoker.test_session import DefaultTestSession
        self.__test_session = DefaultTestSession()
        self.__ref_config = self.__test_session.init(project_root_dir, cli_config, run_id)
        self.__load_console()

        return self.create_test_context(self.__default_context_name)

    def get_logger(self):
        return self.__logger

    def get_console(self):
        return self.__console

    @classmethod
    def get_test_session(cls):
        return cls.__SESSION

    @classmethod
    def has_configuration(cls, config_name):
        return config_name.upper() in cls.thread_map

    @classmethod
    def has_property(cls, config_name, path):
        return cls.has_configuration(config_name) and path.upper() in cls.thread_map[config_name]

    @classmethod
    def merge_configuration(cls, source_config_name, target_config_name):
        if cls.has_configuration(source_config_name):
            if cls.has_configuration(target_config_name):
                cls.thread_map[target_config_name.upper()] = cls.thread_map[source_config_name].clone()
            else:
                cls.register_new_configuration(target_config_name.upper())
                cls.thread_map[target_config_name.upper()] = cls.thread_map[source_config_name.upper()].clone()
        else:
            raise Exception("No source configuration found for name: %s".format(source_config_name))

    @classmethod
    def register_new_configuration(cls, config_name):
        cls.thread_map[config_name.upper()] = {}

    @classmethod
    def update_thread_config(cls, config_name, config):
        cls.thread_map[config_name].clone_add(config)

    @classmethod
    def value(cls, prop_name):
        # uc_prop_name = None
        # if type(prop_name) is not str:
        #     prop_name = self.__get_prop_path_for_enum(prop_name).upper()
        uc_prop_name = prop_name.upper()
        tname = thread_utils.get_current_thread_name()
        if tname in cls.thread_map and uc_prop_name in cls.thread_map[tname]:
            return cls.thread_map[tname][uc_prop_name].value
        elif uc_prop_name in cls.arjuna_options:
            return cls.arjuna_options[uc_prop_name].value
        else:
            raise Exception("No property configred for name: {}".format(prop_name))

    @classmethod
    def get_configured_name(self, section_name, internal_name):
        return self.strings_manager.get_configured_name(section_name, internal_name)

    @classmethod
    def get_problem_text(self, problem_code):
        return self.strings_manager.get_problem_text(problem_code)

    @classmethod
    def get_info_message_text(self, code):
        return self.strings_manager.get_info_message_text(code)

    @classmethod
    def clone_evars(self):
        return copy.deepcopy(self.exec_var_map)

    @classmethod
    def clone_user_options(self):
        return copy.deepcopy(self.user_options)

    def __load_console(self):
        from arjuna.core.enums import ArjunaOption
        dl = logging.getLevelName(self.__ref_config.get_arjuna_option_value(ArjunaOption.LOG_CONSOLE_LEVEL).as_str().upper())
        log_dir = self.__ref_config.get_arjuna_option_value(ArjunaOption.LOG_DIR).as_str()
        if not os.path.isdir(log_dir):
            os.makedirs(log_dir)
        fl = logging.getLevelName(self.__ref_config.get_arjuna_option_value(ArjunaOption.LOG_FILE_LEVEL).as_str().upper())
        fname = self.__ref_config.get_arjuna_option_value(ArjunaOption.PYTHON_LOG_NAME).as_str()
        lpath = os.path.join(log_dir, fname)

        logger = logging.getLogger(self.prog)
        logger.setLevel(logging.DEBUG)
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(dl)
        fh = logging.FileHandler(lpath, "w", 'utf-8')
        fh.setLevel(fl)
        f_fmt = logging.Formatter(u'[%(levelname)5s]\t%(asctime)s\t%(pathname)s::%(module)s.%(funcName)s:%(lineno)d\t%(message)s')
        c_fmt = logging.Formatter(u'[%(levelname)5s]\t%(message)s')
        ch.setFormatter(c_fmt)
        fh.setFormatter(f_fmt)
        logger.addHandler(ch)
        logger.addHandler(fh)

        class __console:
            lock = threading.RLock()

            def __init__(self):
                # self.lock = threading.RLock()
                self.separator = os.linesep
                self.log_display_level = dl
                self.logger = logger

            @sync_method('lock')
            def __log(self, message, err=False):
                if err:
                    mparts = message.replace(u"\r\n", u"--|--").replace(u"\n", u"--|--").split(u'--|--')
                    for mpart in mparts:
                        self.logger.error(mpart)
                else:
                    mparts = message.replace(u"\r\n", u"--|--").replace(u"\n", u"--|--").split(u'--|--')
                    for mpart in mparts:
                        self.logger.info(mpart)

            @sync_method('lock')
            def __print(self, message):
                print(message, end='', file=sys.stdout, flush=True)

            @sync_method('lock')
            def __eprint(self, message):
                print(message, end='', file=sys.stderr, flush=True)

            @sync_method('lock')
            def __println(self, message):
                print(message, file=sys.stdout, flush=True)

            @sync_method('lock')
            def __eprintln(self, message):
                print(message, file=sys.stderr, flush=True)

            @sync_method('lock')
            def __msg(self, *messages):
                return " ".join([str(m) for m in messages])

            @sync_method('lock')
            def display(self, *messages):
                message = self.__msg(*messages)
                should_print = self.__log(message)
                if should_print:
                    self.__println(message)

            @sync_method('lock')
            def display_error(self, *messages):
                message = self.__msg(*messages)
                should_print = self.__log(message, err=True)
                if should_print:
                    self.__eprintln(message)

            @sync_method('lock')
            def error_for_console(self, *messages):
                message = self.__msg(*messages)
                self.__eprintln(message)

            @sync_method('lock')
            def display_on_same_line(self, *messages):
                message = self.__msg(*messages)
                should_print = self.__log(message)
                if should_print:
                    self.__print(message)

            @sync_method('lock')
            def marker(self, length, symbol='-'):
                self.display(symbol * length)

            @sync_method('lock')
            def marker_error(self, length, symbol='-'):
                self.display_error(symbol * length)

            @sync_method('lock')
            def marker_on_same_line(self, length=40):
                self.display_on_same_line("-" * length)

            @sync_method('lock')
            def display_key_value(self, key, value):
                message = "%s %s".format(key, value)
                self.display(message)

            @sync_method('lock')
            def __get_formatted_key_value(self, key, value, left_padding):
                if not left_padding:
                    message = "| {:20s}| {}".format(key, value)
                else:
                    message = "| {}| {}".format(key.ljust(left_padding), value)
                return message

            @sync_method('lock')
            def __display_paddedKV(self, key, value, print_func, left_padding):
                print_func(self.__get_formatted_key_value(key, value, left_padding))

            @sync_method('lock')
            def display_padded_key_value(self, key, value, left_padding=None):
                self.__display_paddedKV(key, value, self.display, left_padding)

            @sync_method('lock')
            def display_padded_key_value_error(self, key, value, left_padding=None):
                self.__display_paddedKV(key, value, self.display_error, left_padding)

            @sync_method('lock')
            def display_exception_block(self, e, strace):
                self.marker_error(80)
                self.display_padded_key_value_error("Exception Type", e.__class__.__name__, 30)
                self.display_padded_key_value_error("Exception Message", str(e), 30)
                self.display_padded_key_value_exception_trace("Exception Trace", strace, 30)
                self.marker_error(80)

            def set_central_log_level(self, level):
                self.central_log_level = level

            @sync_method('lock')
            def display_multiline_key_value(self, key, value, left_padding=30):
                value = str(value)
                ctrace_parts = value.replace("\t", " ").replace("\r\n\r\n","\r\n").replace("\n\n","\n").split(sys_utils.get_line_separator())

                header = self.__get_formatted_key_value(key, ctrace_parts[0], left_padding)
                self.__log(header)
                # if should_print:
                #     self.error_for_console(header)

                for s in ctrace_parts[1:]:
                    message = self.__get_formatted_key_value("", s, left_padding)
                    self.__log(message)
                    # if should_print:
                    #     self.(message)

            display_padded_key_value_exception_trace = display_multiline_key_value
        self.__logger = logger
        self.__console = __console()

    def get_central_config(self):
        return self.__ref_config

    def create_test_context(self, name):
        '''
            Creates test context object.
            ?? Does it take central config??
        '''

        from arjuna.configure.invoker.context import DefaultTestContext
        context = DefaultTestContext(self.__test_session, name)
        self.__context_map[name] = context
        return context

class Arjuna:
    '''
        Facade of Arjuna framework.
        Contains static methods which wrapper an internal singleton class for easy access to top level Arjuna functions.
    '''

    ARJUNA_SINGLETON = None
    LOGGER = None

    @classmethod
    def init(cls, project_root_dir, cli_config=None, run_id=None):
        '''
            Returns reference test context which contains reference configuration.
            This reference test context merges central conf, project conf and central CLI options.
            Root directory is assumed as per the project structure.
            You can also provide an alternative root directory for test project.
        '''
        cls.ARJUNA_SINGLETON = ArjunaSingleton()
        return cls.ARJUNA_SINGLETON.init(project_root_dir, cli_config, run_id)

    @classmethod
    def init_logger(cls, testsession_id, log_dir):
        logger = logging.getLogger("setu")
        logger.setLevel(logging.DEBUG)
        # ch = logging.StreamHandler(sys.stdout)
        # ch.setLevel(logging.INFO)
        fh = logging.FileHandler(log_dir + "/arjuna-setu-{}-ts-{}.log".format(time.time(), testsession_id), "w", 'utf-8')
        fh.setLevel(logging.DEBUG)
        f_fmt = logging.Formatter(u'[%(levelname)5s]\t%(asctime)s\t%(pathname)s::%(module)s.%(funcName)s:%(lineno)d\t%(message)s')
        c_fmt = logging.Formatter(u'[%(levelname)5s]\t%(message)s')
        # ch.setFormatter(c_fmt)
        fh.setFormatter(f_fmt)
        # logger.addHandler(ch)
        logger.addHandler(fh)
        cls.LOGGER = logger

    @classmethod
    def get_logger(cls):
        '''
            Returns framework logger.
        '''
        return cls.ARJUNA_SINGLETON.get_logger()

    @classmethod
    def get_console(cls):
        '''
            Returns framework console.
        '''
        return cls.ARJUNA_SINGLETON.get_console()

    @classmethod
    def get_ref_config(cls):
        '''
            Returns the reference configuration.
        '''
        return cls.ARJUNA_SINGLETON.get_central_config()

    @classmethod
    def get_gui_mgr(cls):
        '''
            Returns the central GUI Manager object that mangages namespaces.
        '''
        return cls.ARJUNA_SINGLETON.gui_mgr

    @classmethod
    def get_ref_arjuna_option_value(cls, option):
        return cls.get_central_config().get_ref_arjuna_option_value(option)

    @classmethod
    def get_ref_user_option_value(cls, option):
        return cls.get_central_config().get_ref_user_option_value(option) 

    @staticmethod
    def get_test_context(name):
        '''
            Returns test context for a given name (case-insensitive).
        '''
        pass

    @staticmethod
    def register_test_context(context):
        '''
            Registers a test context object.
        '''
        pass    

    @classmethod
    def create_test_context(cls, name):
        '''
            Creates test context object.
            ?? Does it take central config??
        '''
        return cls.ARJUNA_SINGLETON.create_test_context(name)

    @staticmethod
    def get_root_dir(context):
        '''
            Returns root directory of test project.
        '''
        pass

    @staticmethod
    def create_datasource_builder():
        '''
            Creates and returns DataSourceBuilder object.
        '''
        pass

    @staticmethod
    def exit():
        '''
            Clean-up and finalise resources currently opened by Arjuna.
        '''
        pass


from arjuna.interact.gui.auto.base.configurable import Configurable

class GuiWebRadioGroup(Configurable):

    def __init__(self, gui, lmd, parent=None):
        super().__init__(gui)
        self.__gui = gui
        self.__automator = gui.automator
        self.__finder = parent and parent or self.automator
        self.__radios = None
        self.__find(self.__finder, lmd)

    @property
    def gui(self):
        return self.__gui

    @property
    def automator(self):
        return self.__automator

    def __validate_radio_buttons(self, source):
        if [t for t in source.get_tag_names() if t.strip().lower() != 'input']:
            raise Exception("Not a valid radio group. Contains non-input elements.")
        if [t for t in source.get_attr_values("type") if t.strip().lower() != 'radio']:
            raise Exception("Not a valid radio group. Contains non-radio elements.")
        names = source.get_attr_values("name")
        if len(set(names)) != 1:
            raise Exception("Not a valid radio group. Contains radio elements belonging to different radio groups.")

    def __check_type_if_configured(self, tags):
        if self._should_check_type(): self.__validate_radio_buttons(tags)

    def __find(self, finder, lmd):
        # This would force the identification of partial elements in the wrapped multi-element.
        self.__radios = finder.multi_element(self.gui, lmd)
        self.__check_type_if_configured(self.source)
        self.__radios.configure_partial_elements(self.settings)

    def has_index_selected(self, index):
        return self.__radios[index].is_selected()

    def has_value_selected(self, value):
        return self.__radios.get_instance_by_value(value).is_selected()

    @property
    def value(self):
        instance = self.__radios.get_first_selected_instance()
        return instance.source.get_attr_value("value")

    def __select_option(self, option):
        option.select()
        if self._should_check_post_state() and not option.is_selected():
            raise Exception("The attempt to select the radio button was not successful.")

    def select_index(self, index):
        option = self.__radios[index]
        self.__select_option(option)

    def select_ordinal(self, ordinal):
        return self.select_by_index(ordinal-1)

    def select_value(self, value):
        option = self.__radios.get_instance_by_value(value)
        self.__select_option(option)

    @property
    def source(self):
        return self.__radios.source


import unittest

class TestRunner:
    import sys
    import os
    import unittest
    
    def __init__(self):
        from arjuna import Arjuna
        from arjuna.core.enums import ArjunaOption
        self.__project_dir = Arjuna.get_ref_config().get_arjuna_option_value(ArjunaOption.PROJECT_ROOT_DIR).as_str()
        self.__tests_dir = Arjuna.get_ref_config().get_arjuna_option_value(ArjunaOption.PROJECT_TESTS_DIR).as_str()
        self.__suite = unittest.TestSuite()

    @property
    def tests_dir(self):
        return self.__tests_dir

    @property
    def test_suite(self):
        return self.__suite
    
    def load_all_tests(self):
        test_loader = unittest.TestLoader()
        suite = test_loader.discover(self.tests_dir)
        self.test_suite.addTest(suite)
        # from tests import fleet
        # suite.addTest(loader.loadTestsFromModule(fleet))
    
    def run(self):
        runner = unittest.TextTestRunner(verbosity=3)
        runner.run(self.test_suite)


import re
import os
from lxml import etree, html
from io import StringIO
from collections import namedtuple

def process_child_html(in_str):
    processed = os.linesep.join([l for l in in_str.splitlines() if l.strip()])
    return "\t" + processed

def remove_empty_lines_from_string(in_str):
    return ' '.join([l.strip() for l in in_str.splitlines() if l.strip()])

def empty_or_none(in_str):
    if type(in_str) is str and not in_str.strip():
        return True
    else:
        return in_str is None

SourceContent = namedtuple('SourceContet', "all root inner text")

class ElementXMLSourceParser:

    def __init__(self, raw_source, root_element="body"):
        super().__init__()
        self.__raw_source = raw_source
        self.__root_element = root_element
        self.__fpaths = []
        self.__content = None

    @property
    def content(self):
        return self.__content

    def load(self):
        raw_source = self.__raw_source
        parser = etree.HTMLParser(remove_comments=True)
        tree = etree.parse(StringIO(raw_source), parser)
        if self.__root_element == "body":
            body = tree.getroot().find('body')
            elem_node = list(body)[0]
        else:
            body = tree.getroot()
            elem_node = body

        # print(etree.tostring(elem_node, encoding='unicode'))
        non_normalized_text_content = os.linesep.join(
            [
                remove_empty_lines_from_string(c.text)
                for c in elem_node.iter() 
                if not empty_or_none(c.text)
            ]).strip()

        non_normalized_inner_html = os.linesep.join([
            process_child_html(etree.tostring(c, encoding='unicode'))
            for c in list(elem_node.iterchildren())
            ])

        self.__tag = elem_node.tag
        self.__text_content = non_normalized_text_content
        self.__inner_html = non_normalized_inner_html
        self.__raw_attrs = elem_node.keys()
        self.__attributes = {k.lower():v for k,v in elem_node.items()}
        self.__full_source = raw_source

        for child in list(elem_node): elem_node.remove(child)
        self.__self_source = ' '.join(etree.tostring(elem_node, encoding=str).splitlines())
        self.__self_source = re.sub(r"\s+", " ", self.__self_source)

        self.__content = SourceContent(all=self.__full_source, root=self.__self_source, inner=self.__inner_html, text=self.__text_content)

    @property
    def tag(self):
        return self.__tag

    def get_attr_names(self):
        return self.__raw_attrs

    def is_attr_present(self, attr):
        return attr in self.__attributes.keys()

    def get_attr_value(self, attr_name, optional=False):
        try:
            return self.__attributes[attr_name.lower()]
        except Exception as e:
            if optional:
                return None
            else:
                raise Exception("Attribute {} not found for element".format(attr_name))

    def get_root_content(self):
        return self.__self_source

    def get_inner_content(self):
        return self.__inner_html

    def get_full_content(self):
        return self.__full_source

    def get_text_content(self):
        return self.__text_content

class MultiElementSource:

    def __init__(self):
        super().__init__()
        self.__instances = None

    def load(self, instances):
        self.__instances = instances

    def get_full_content(self):
        return os.linesep.join([e.source.content.all for e in self.__instances])

    def get_inner_content(self):
        return os.linesep.join([e.source.content.inner for e in self.__instances])

    def get_text_content(self):
        return os.linesep.join([e.source.content.root for e in self.__instances])

    def get_root_content(self):
        return os.linesep.join([e.source.content.root for e in self.__instances])

    def _get_root_content_as_list(self):
        return [e.source.content.root for e in self.__instances]

    def get_tag_names(self):
        return [e.source.tag for e in self.__instances]

    def get_text_contents(self):
        return [e.source.get_text_content() for e in self.__instances]

    def get_values(self):
        return [e.source.get_attr_value("value") for e in self.__instances]

    def get_attr_values(self, attr):
        return [e.source.get_attr_value(attr) for e in self.__instances]

class FrameSource:

    def __init__(self, frame):
        super().__init__()
        self.__frame = frame
        self.__root_source = None
        self.__html_source = None

    def set_root_source(self, src):
        '''
        Once frame switch takes place, this source can not be got. Hence needs to happen
        explicitly at the time of wrapped element finding.
        '''
        self.__root_source = src

    def load(self):
        self.__frame.focus()
        self.__html_source = self.__frame._get_html_content_from_remote()

    def get_full_content(self):
        return self.get_root_content() + self.get_inner_content()

    def get_inner_content(self):
        return self.__html_source.get_full_content()

    def get_text_content(self):
        return self.__html_source.get_text_content()

    def get_root_content(self):
        return self.__root_source




import requests
import json
import os

from arjuna.setu import Setu


class SetuAgentRequester:

    def __init__(self, base_url):
        self.base_url = base_url

    def __convert_response_text_to_json(self, response_text):
        try:
            return json.loads(response_text)
        except:
            raise Exception("Response text is not JSON. Raw response text: " + response_text)

    def __raise_exception_if_error(self, response_dict):
        if response_dict is None:
            raise Exception("Setu Actor's JSON response is null.")
        if (response_dict["result"].lower() == "error"):
            emessage = response_dict.get("emessage", "")
            raise Exception("Agent Error: " + emessage + os.linesep + "Trace: " + response_dict["etrace"])

    def get(self, url):
        req_url = self.base_url + url
        response = requests.get(req_url)
        response_json = json.loads(response.text)
        self.__raise_exception_if_error(response_json)
        return json.loads(response.text, encoding="utf-8")

    def post(self, url, json_dict):
        req_url = self.base_url + url
        response = requests.post(req_url, json=json_dict)
        response_json = self.__convert_response_text_to_json(response.text)
        self.__raise_exception_if_error(response_json)
        return json.loads(response.text)

    def get_base_url(self):
        return self.base_url


import json
from enum import Enum

from arjuna.configure.impl.processor import ConfigCreator, CentralConfigLoader, ProjectConfigCreator
from arjuna.core.reader.hocon import HoconStringReader, HoconConfigDictReader
from arjuna.core.value import AnyRefValue

class TestConfigurator:

    def __init__(self):
        self.__default_ref_config = None
        self.__config_map = {}
        self.__cli_central_config = None
        self.__cli_test_config = None
        self.__project_config_loaded = False

    def __create_config_from_option_dicts(self, reference, arjuna_options, user_options):
        def format_value(val):
            if isinstance(val, AnyRefValue):
                return val.as_str()
            elif isinstance(val, Enum):
                return val.name
            else:
                return str(val)

        # ANCHOR
        crawdict = {
            "arjunaOptions": {k:format_value(v) for k,v in arjuna_options.items()},
            "userOptions": {k:format_value(v) for k,v in user_options.items()}
        }
        hreader = HoconStringReader(json.dumps(crawdict))
        hreader.process()

        config = ConfigCreator.create_new_conf(
            self.__default_ref_config.processor,
            reference,
            hreader.get_map()
        )
        return config

    def __init_cli_dicts(self, arjunaCentralOptions={}, arjunaTestOptions={}, userCentralOptions={}, userTestOptions={}):
        self.__cli_central_config = self.__create_config_from_option_dicts(None, arjunaCentralOptions, userCentralOptions)
        self.__cli_test_config = self.__create_config_from_option_dicts(None, arjunaTestOptions,
                                                                        userTestOptions)

    def init(self, root_dir, cli_config):
        self.__default_ref_config = CentralConfigLoader(root_dir).config
        # self.__config_map[self.__default_ref_config.id] = self.__default_ref_config
        cli_config = cli_config and cli_config or {}
        self.__init_cli_dicts(**cli_config)
        return self.__default_ref_config

    def create_project_conf(self):
        project_conf_loader = ProjectConfigCreator(self.__default_ref_config)
        self.__default_ref_config = project_conf_loader.config
        self.__default_ref_config.process_arjuna_options()
        self.__default_ref_config.update(self.__cli_central_config)
        self.__default_ref_config.update(self.__cli_test_config)
        # self.__config_map[self.__default_ref_config.id] = self.__default_ref_config
        return self.__default_ref_config

    @property
    def ref_config(self):
        return self.__default_ref_config

    def get_config(self, id):
        return self.__config_map[id]

    # def get_arjuna_option_value(self, config_setu_id, option):
    #     sname = ArjunaOption[option.upper().strip().replace(".", "_")]
    #     rvalue = self.__config_map[config_setu_id].get_arjuna_option_value(sname)
    #     return rvalue

    # def get_user_option_value(self, config_setu_id, option):
    #     sname = option.upper().strip().replace(".", "_")
    #     rvalue = self.__config_map[config_setu_id].user_config.value(sname)
    #     return rvalue

    # def get_central_arjuna_option_value(self, option):
    #     sname = ArjunaOption[option.upper().strip().replace(".", "_")]
    #     rvalue = self.__default_ref_config.get_arjuna_option_value(sname)
    #     return rvalue

    # def get_central_user_option_value(self, option):
    #     sname = option.upper().strip().replace(".", "_")
    #     rvalue = self.__default_ref_config.user_config.value(sname)
    #     return rvalue

    def register_new_config(self, arjuna_options, user_options, parent_config=None):
        # Registering a config is post project conf registration. If no project conf, set it to true.
        self.__project_config_loaded = True
        reference = parent_config and parent_config._wrapped_config or self.__default_ref_config._wrapped_config
        config = self.__create_config_from_option_dicts(reference, arjuna_options, user_options)
        config.update(self.__cli_test_config)
        config.process_arjuna_options()
        return config



# Skeleton. needs to be ported.

class ArjunaSingleton:

    def __init__(self):
        self.__DEFAULT_CONTEXT_NAME = "default_context"
        self.__root_dir = None
        self.__test_session = test_session
        self.__ref_config = None
        self.__cli_config = None
        self.__logger = None
        self.__test_context_map = dict()

    def init(self, root_dir=None):
        self.__root_dir = root_dir
        self.__cli_config = CliArgsConfig()
        # Finalize logger
        createLogger("arjuna", rootDir + File.separator + "log" + File.separator + "arjuna-java.log")
        logger = Logger.getLogger("arjuna")
        Console.init()

        self.__test_session = DefaultTestSession()
        self.__ref_config = self.__test_session.init(root_dir)

        context = this.createTestContext(self.__DEFAULT_CONTEXT_NAME)
        self.__test_session[self.__DEFAULT_CONTEXT_NAME] = context
        return context

    @property
    def cli_config(self):
        return self.__cli_config

    @property
    def ref_config(self):
        return self.__ref_config

    @property
    def root_dir(self):
        return self.__root_dir

    def register_test_context(self, context):
        self.__test_context_map[context.name.lower()] = context

    def get_test_context(name):
        if name is None:
            raise Exception("Context name was passed as None.")
        try:
            return self.__test_context_map[name.lower()]
        except:
            raise Exception("No context found with name: " + name)

    def create_test_context(name):
        return DefaultTestContext(self.__test_session, name)

    def normalize_user_option(self, option):
        return option.strip().upper().replace(".", "_")

    def normalize_arjuna_option(self, option):
        return ArjunaOption[self.normalize_user_option(option)]

    def create_data_source_builder(self):
        return DataSourceBuilder(self.__test_session)

    def create_gui_automator(self, config=None, extended_config=None):
        return DefaultGuiAutomator(config and config or self.__ref_config, extended_config)

    def create_logger(self, logger_name, log_file_path):
        pass


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import abc
import pprint
from enum import Enum
from .types import *
from functools import partial

from weakref import WeakKeyDictionary


class EnumSwitch:
    def __init__(self, cases, const_args):
        self.dispatch_dict = cases
        self.const_args = const_args

    def __call__(self, econst):
        if econst in self.dispatch_dict:
            f_tuple = self.dispatch_dict[econst]
            arg_list = list(self.const_args) + list(f_tuple[1:])
            return f_tuple[0](*arg_list)



class PropertyEnum(Enum):
    pass
    # def _generate_next_value_(name, start, count, last_values):
    #     return ".".join(name.lower().split("_"))


class Descriptor(metaclass=abc.ABCMeta):
    def __init__(self):
        self._name = None
        self._values = WeakKeyDictionary()

    def __get__(self, instance, cls):
        return self._values.get(instance)

    def __set__(self, instance, value):
        self.validate(instance, value)
        if instance in self._values:
            self._values[instance].set_value(value)
        else:
            self._values[instance] = self._create_value(value)

    def set_name(self, name):
        self._name = name

    @abc.abstractmethod
    def _create_value(self, value):
        pass

    def validate(self, instance, value):
        self._values[instance].validate(value)

class Decorator(metaclass=abc.ABCMeta):
    def __new__(cls, *args, **kwargs):
        if cls.__name__ == "Decorator":
            raise Exception("Decorator is an abstract class. Inherit to implement a concrete decorator.")
        else:
            return super().__new__(cls)

    def __init__(self, **dkwargs):
        super().__init__()
        self.__args = CIStringAttrDict(dkwargs)
        self.__kallable = None
        return self

    def __call__(self, kallable):
        self.__kallable = kallable
        return partial(self._decorated, self.__args, self.__kallable)

    @abc.abstractmethod
    def _decorated(self, dargs, kallable, *fvargs, **fkwargs):
        pass

# helpers
def dynamic_load(global_dict, all_list, names):
    for n, b, d in names:
        global_dict[n] = type(n, b, d)
        all_list.append(n)

def prepare_locals(l):
    return {i: j for i, j in l.items() if i not in ['self', '__class__']}



'''
import logging
import sys
import inspect
import time

from arjuna.interact.gui.auto.automator.guiautomator import GuiAutomator
from arjuna.interact.gui.auto.finder.emd import GuiElementMetaData, SimpleGuiElementMetaData
from functools import partial

class _DummyLogger:

    def dummy(self, item, *vargs):
        print("Dummy call", item)
        frame = inspect.stack()[3]
        caller_str = "{}.py:{}:L{}".format(inspect.getmodule(frame[0]).__name__, frame[3], frame[2])
        print("Dummy logger called by {} with args: ".format(caller_str), str(vargs)[:300], "...")
        pass

    def __getattr__(self, item):
        return partial(self.dummy, item)


class Setu:
    LOGGER = None
    __DUMMY_LOGGER = _DummyLogger()

    @classmethod


    @classmethod
    def get_logger(cls):
        return cls.LOGGER and cls.LOGGER or cls.__DUMMY_LOGGER


__all__ = ["Setu", "GuiAutomator", "GuiElementMetaData", "SimpleGuiElementMetaData"]
'''


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import xlrd


class ExcelRowReader:
    def __init__(self, path):
        self.wb = xlrd.open_workbook(path)
        self.sheet = self.wb.sheet_by_name(self.wb.sheet_names()[0])
        self.rcount = self.sheet.nrows
        self.ccount = self.sheet.ncols
        self.curent_row_index = -1
        self.validate()

    def __iter__(self):
        return self

    def read_next_row(self):
        self.curent_row_index += 1
        if self.curent_row_index < self.rcount:
            return self.sheet.row(self.curent_row_index)
        else:
            raise Exception("Done")

    def next(self):
        try:
            return self.process(self.read_next_row())
        except:
            self.close()
            raise StopIteration()

    def read(self):
        return [r for r in iter(self)]

    def process(self, row):
        return row

    def validate(self):
        pass

    def close(self):
        pass
        #self.wb.close()


class ExcelRow2ArrayReader(ExcelRowReader):
    def __init__(self, path):
        super().__init__(path)
        self.headers = []
        self._populate_headers()

    def get_headers(self):
        return self.headers

    def _populate_headers(self):
        self.headers = [h.value for h in self.read_next_row()]

    def process(self, row):
        return [h.value for h in row]

    def validate(self):
        if self.rcount == 0:
            raise Exception("Empty or wrongly formattted Excel file. Is first line empty?")


class ExcelRow2MapReader(ExcelRow2ArrayReader):
    def __init__(self, path):
        super().__init__(path)

    def process(self, row):
        return zip(self.headers, super().process(row))



from arjuna.interact.gui.auto.element.guielement import GuiElement
from arjuna.core.enums import ArjunaOption

class WebAlert:

    def __init__(self, automator):
        self.__automator = automator

    @property
    def automator(self):
        return self.__automator

    def confirm(self):
        self.automator.dispatcher.confirm_web_alert()
        self.automator.alert_handler.delete_alert()

    def dismiss(self):
        self.automator.dispatcher.dismiss_web_alert()
        self.automator.alert_handler.delete_alert()

    @property
    def text(self):
        return self.automator.dispatcher.get_text_from_web_alert()

    @text.setter
    def text(self, text):
        self.automator.dispatcher.send_text_to_web_alert(text)


from .driverelement import SeleniumDriverElementDispatcher
from arjuna.interact.gui.dispatcher.driver.driver_commands import DriverCommands
from arjuna.interact.gui.dispatcher.driver.element_finder import ElementFinder
from arjuna.interact.gui.dispatcher.driver.melement import MultiElement
from selenium.webdriver.remote.webelement import WebElement

class SeleniumDriverDispatcher:

    def __init__(self):
        self.__config = None
        self.__driver = None

    def __create_gui_element_dispatcher(self, element):
        return SeleniumDriverElementDispatcher.create_dispatcher(self, element)

    @property
    def driver(self):
        return self.__driver

    def launch(self, config):
        self.__config = config
        from .browser_launcher import BrowserLauncher
        self.__driver = BrowserLauncher.launch(config) 

    def quit(self):
        DriverCommands.quit(self.__driver)

    def go_to_url(self, url):
        DriverCommands.go_to_url(self.__driver, url)

    def go_back_in_browser(self):
        DriverCommands.go_back_in_browser(self.__driver)

    def go_forward_in_browser(self):
        DriverCommands.go_forward_in_browser(self.__driver)

    def refersh_browser(self):
        DriverCommands.refersh_browser(self.__driver)

    def get_source(self):
        return DriverCommands.get_source(self.__driver)

    def execute_javascript(self, script):
        return DriverCommands.execute_javascript(self.__driver, script)

    def take_screenshot(self):
        DriverCommands.take_screenshot(self.__driver)

    def find_element(self, with_type, with_value):
        element = ElementFinder.find_element(self.__driver, with_type, with_value)
        return 1, self.__create_gui_element_dispatcher(element)

    def __process_single_js_element(self, element):
        # JS returns null, undefined
        if element is None:
            raise Exception("JavaScript could not find element.")
        elif not isinstance(element, WebElement):
            raise Exception("JavaScript returned a non-element object.")
        else:
            return element

    def __process_js_element_list(self, elements):
        if not elements: raise Exception("JavaScript could not find element.")
        return [self.__process_single_js_element(e) for e in elements]

    def __process_js_element(self, element):
        if type(element) is list:
            element = self.__process_js_element_list(element)[0]
        else:
            element = self.__process_single_js_element(element)
        return element

    def __process_js_multielement(self, elements):
        if type(elements) is list:
            elements = self.__process_js_element_list(elements)
        else:
            elements = [self.__process_single_js_element(elements)]
        return elements

    def find_element_with_js(self, js):
        element = self.execute_javascript(js)
        element = self.__process_js_element(element)
        return 1, self.__create_gui_element_dispatcher(element)

    def find_multielement(self, with_type, with_value):
        web_elements = ElementFinder.find_elements(self.__driver, with_type, with_value)
        melement = MultiElement([SeleniumDriverElementDispatcher(self, web_element) for web_element in web_elements])
        return melement.get_instance_count(), melement

    def find_multielement_with_js(self, js):
        web_elements = self.execute_javascript(js)
        web_elements = self.__process_js_multielement(web_elements)
        melement = MultiElement([SeleniumDriverElementDispatcher(self, web_element) for web_element in web_elements])
        return melement.get_instance_count(), melement

    def get_current_window_handle(self):
        return DriverCommands.get_current_window_handle(self.__driver)

    def get_current_window_title(self):
        return DriverCommands.get_window_title(self.__driver)

    def maximize_current_window(self):
        DriverCommands.maximize_window(self.__driver)

    def get_current_window_size(self):
        res = DriverCommands.get_current_window_size(self.__driver)
        return {"width" : res[0], "height" : res[1]}

    def get_all_window_handles(self):
        return DriverCommands.get_all_winodw_handles(self.__driver)

    def focus_on_window(self, handle):
        DriverCommands.focus_on_window(self.__driver, handle)

    def close_current_window(self):
        DriverCommands.close_current_window(self.__driver)

    def is_web_alert_present(self):
        return DriverCommands.is_web_alert_present(self.__driver)

    def confirm_web_alert(self):
        DriverCommands.confirm_web_alert(self.__driver)

    def dismiss_web_alert(self):
        DriverCommands.dismiss_web_alert(self.__driver)

    def get_text_from_web_alert(self):
        return DriverCommands.get_text_from_web_alert(self.__driver)

    def send_text_to_web_alert(self, text):
        DriverCommands.send_text_to_web_alert(self.__driver, text)

    def focus_on_frame(self, element_dispatcher):
        DriverCommands.focus_on_frame(self.__driver, element_dispatcher.driver_element)

    def get_element_for_setu_id(self, id):
        return self.__driver_elements[id]

    def focus_on_parent_frame(self):
        DriverCommands.focus_on_parent_frame(self.__driver)

    def focus_on_dom_root(self):
        DriverCommands.focus_on_dom_root(self.__driver)

    def perform_action_chain(self, action_chain):
        DriverCommands.perform_action_chain(self, self.__driver, action_chain)

    def hover_on_element(self, element_dispatcher):
        DriverCommands.hover_on_element(self.__driver, element_dispatcher.driver_element)






from arjuna.core.enums import ArjunaOption

class Browser:

    def __init__(self, automator):
        self.__automator = automator

    @property
    def automator(self):
        return self.__automator

    def go_to_url(self, url):
        self.automator.dispatcher.go_to_url(url=url)

    def go_back(self, url):
        self.automator.dispatcher.go_back()

    def go_forward(self, url):
        self.automator.dispatcher.go_forward()

    def refresh(self, url):
        self.automator.dispatcher.refresh()

    def execute_javascript(self, js):
        self.automator.dispatcher.execute_javascript(script=js)


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import abc
import inspect
from numbers import Number as PyNumber
from weakref import WeakKeyDictionary
from enum import Enum

from . import constants


def _get_cstr(context=None):
    return context and "Encountered Type error for {}.".format(context) or ""


class TypeValidationError(TypeError):
    pass


class Descriptor:
    pass

class String(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if String.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a string.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return isinstance(input, str)

class Integer(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if Integer.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a integer.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return isinstance(input, str)

class Float(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if Float.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a flot.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return isinstance(input, str)

class Number(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if Number.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a number.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return isinstance(input, PyNumber)

class Bool(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if Bool.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a bool.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return isinstance(input, bool)

    @staticmethod
    def force_convert(input):
        if Bool.check(input):
            return input
        else:
            if input in constants.TRUES:
                return True
            elif input in constants.FALSES:
                return False
            else:
                return bool(input)

class Enumeration(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if Enumeration.check(input):
            raise TypeValidationError("{} Provided enum_type: {} [type: {}] is not a class.".format(_get_cstr(context), input, type(input)))
        return input

    @staticmethod
    def check(input):
        if not inspect.isclass(input) or not issubclass(input, Enum):
            return False

class EnumConstant(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if EnumConstant.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a enum.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return isinstance(input, Enum)

    @staticmethod
    def convert(enum_type, estr, context=None, enum_type_validated=False):
        if not enum_type_validated: Enumeration.validate(enum_type)

        if String.check(estr):
            uestr = estr.upper()
            if uestr in enum_type.__members__:
                return enum_type[uestr]
            else:
                raise TypeError("{} {} is not a valid enum constant for enum {}.".format(_get_cstr(context), estr, enum_type))

class EnumConstantList(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if EnumConstantList.check(input):
            raise TypeValidationError("{} {} [type: {}] is not an enum constant list.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        if not List.check(input): return False
        for i in input:
            if not EnumConstant.check(i):
                return False
        return True

    @staticmethod
    def convert(enum_type, input, context=None):
        Enumeration.validate(enum_type)
        il = List.check(input) and input or [input]
        return [EnumConstant.convert(enum_type, estr, enum_type_validated=True) for estr in il]

class List(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if not List.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a list.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return isinstance(input, list)

class Tuple(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if not Tuple.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a tuple.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return isinstance(input, tuple)

class Set(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if not Set.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a set.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return isinstance(input, set)

class ListTupleOrSet(Descriptor):

    @staticmethod
    def validate(input, context=None):
        if not ListTupleOrSet.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a list, tuple or set.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return List.check(input) or Tuple.check(input) or Set.check(input)

class StringList(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if not isinstance(input, list):
            raise TypeValidationError("{} {} [type: {}] is not a list.".format(_get_cstr(context)))
        else:
            for i in input:
                String.validate(i, "Validating String List. Provided List: {}.".format(str(input)))
        return input

    @staticmethod
    def force_convert(input):
        if input is None:
            return None
        else:
            if String.check(input):
                return [input]
            elif Number.check(input):
                return [str(input)]
            elif ListTupleOrSet.check(input):
                return [str(i) for i in input]
            else:
                raise TypeError("Can not force convert >>{}<< to a string list.".format(str(input)))

class NaturalNumber(Descriptor):
    @staticmethod
    def validate(input, context=None):
        Integer.validate(input, context)
        if input <= 0:
            raise TypeValidationError("{} {} [type: {}] is not a natural number.".format(_get_cstr(context)))
        return input


class RangeBoundInt(Descriptor):
    @staticmethod
    def validate(input, left, right, context=None):
        Integer.validate(input, context)
        if input <= 0:
            raise TypeValidationError("{} {} [type: {}] is not in the range [{}, {}].".format(_get_cstr(context)))
        return input



import abc
import os
from enum import Enum
import unittest

from arjuna.interact.gui.auto.finder.emd import GuiElementMetaData

from .guidef import *

class AsserterMixIn:

    def __init__(self):
        # Trick to use assertions outside of a unittest test
        self._asserter = unittest.TestCase('__init__')

class Gui(AsserterMixIn):

    def __init__(self, *, config=None, ext_config=None, label=None):
        '''
            You can either provide automator.
        '''
        super().__init__()
        from arjuna import Arjuna
        self.__config = config is not None and config or Arjuna.get_ref_config()
        self.__econfig = ext_config
        if ext_config is None:
            self.__econfig = dict()
        else:
            if type(ext_config) is dict:
                self.__econfig = ext_config
            else:
                self.__econfig = ext_config.config
        self.__label = label is not None and label or self.__class__.__name__

    @property
    def config(self):
        return self.__config

    @property
    def ext_config(self):
        return self.__econfig

    @property
    def label(self):
        return self.__label

    @property
    def name(self):
        return self.__class__.__name__

    @property
    def qual_name(self):
        return self.__class__.__qualname__


class AppPortion(Gui):

    def __init__(self, *, automator, label=None):
        super().__init__(config=automator.config, ext_config=automator.ext_config, label=label)
        self.__app = automator.app
        self.__automator = automator
        from arjuna import Arjuna
        self.__guimgr = Arjuna.get_gui_mgr()
        self.__guidef = None

        self.__gui_registered = False
        self.__def_file_name = None
        self.__def_file_path = None

    @property
    def app(self):
        return self.__app
    
    @property
    def automator(self):
        return self.__automator

    @property
    def gui_def(self):
        return self.__guidef

    def externalize_guidef(self, ns_dir=None, def_file_name=None):
        self.__def_file_name = def_file_name is not None and def_file_name or "{}.gns".format(self.label)        
        from arjuna.core.enums import ArjunaOption
        ns_dir = ns_dir and ns_dir or self.app.ns_dir
        ns_root_dir = self.config.get_arjuna_option_value(ArjunaOption.GUIAUTO_NAMESPACE_DIR).as_str()
        self.__def_file_path = os.path.join(ns_root_dir, ns_dir, self.def_file_name)
        self.__guidef = GuiDef(self.__guimgr.name_store, self.automator, self.label, self.__def_file_path) # self.__guimgr.namespace_dir, 
        # if register:
        #     self._register()

    @property
    def def_file_name(self):
        return self.__def_file_name

    @property
    def def_file_path(self):
        return self.__def_file_path

    def reach_until(self):
        # Children can override and write any necessary loading instructions
        pass

    def validate_readiness(self):
        pass

    def __load(self):
        try:
            self.validate_readiness()
        except:
            try:
                self.reach_until()
                self.validate_readiness()
            except Exception as e:
                raise Exception(
                    "UI [{}] with SetuId [{}] did not load as expected. Error: {}.",
                    self.__class__.__name__,
                    self.get_setu_id(),
                    str(e)
                )

    def transit(self, page):
        pass

    def convert_to_with_lmd(self, *raw_str_or_with_locators):
        from arjuna.interact.gui.helpers import With, WithType
        out = []
        for locator in raw_str_or_with_locators:
            w = None
            if isinstance(locator, With):
                w = locator
            elif type(locator) is str:
                w = With.gns_name(locator)
            elif isinstance(locator, Enum):
                w = With.gns_name(locator.name)
            else:
                raise Exception("A With object or name of element is expected as argument.")

            if w.wtype == WithType.GNS_NAME:
                out.extend(self.gui_def.convert_to_with(w))
            else:
                out.append(w)
        lmd = GuiElementMetaData.create_lmd(*out)
        return lmd

    @property
    def browser(self):
        return self.impl_gui.browser

    def element(self, *str_or_with_locators, iconfig=None):
        return self.automator.element(self, self.convert_to_with_lmd(*str_or_with_locators), iconfig=iconfig)

    def multi_element(self, *str_or_with_locators, iconfig=None):
        return self.automator.multi_element(self, self.convert_to_with_lmd(*str_or_with_locators), iconfig=iconfig)

    def dropdown(self, *str_or_with_locators, option_container_locator=None, option_locator=None, iconfig=None):
        return self.automator.dropdown(
            self, 
            self.convert_to_with_lmd(*str_or_with_locators),
            option_container_lmd=option_container_locator and self.convert_to_with_lmd(option_container_locator) or None,
            option_lmd=option_locator and self.convert_to_with_lmd(option_locator) or None,
            iconfig=iconfig
        )

    def radio_group(self, *str_or_with_locators, iconfig=None):
        return self.automator.radio_group(self, self.convert_to_with_lmd(*str_or_with_locators), iconfig=iconfig)

    def tab_group(self, *str_or_with_locators, tab_header_locator, content_relation_attr, content_relation_type, iconfig=None):
        return self.automator.tab_group(
            self,
            self.convert_to_with_lmd(*str_or_with_locators),
            tab_header_lmd=self.convert_to_with_lmd(tab_header_locator),
            content_relation_attr=content_relation_attr, 
            content_relation_type=content_relation_type, 
            iconfig=iconfig
        )

    @property
    def dom_root(self):
        return self.automator.dom_root(self)

    def frame(self, *str_or_with_locators, iconfig=None):
        return self.automator.frame(self, self.convert_to_with_lmd(*str_or_with_locators), iconfig=iconfig)

    @property
    def alert(self):
        return self.automator.alert

    @property
    def main_window(self):
        return self.automator.main_window

    def child_window(self, *str_or_with_locators):
        return self.automator.child_window(self.convert_to_with_lmd(*str_or_with_locators))

    @property
    def latest_child_window(self):
        return self.automator.latest_child_window

    def close_all_child_windows(self):
        self.automator.close_all_child_windows()

    @property
    def browser(self):
        return self.automator.browser

    def set_slomo(self, on, interval=None):
        self.automator.set_slomo(on, interval)

    def execute_javascript(self, js):
        return self.automator.execute_javascript(js)


import base64
import os
import time

from arjuna.core.enums import ArjunaOption
from arjuna.interact.gui.auto.base.container import ElementContainer
from arjuna.interact.gui.auto.base.dispatchable import Dispatchable
from .drivercaps import DriverCapabilities
from arjuna.interact.gui.auto.source.parser import ElementXMLSourceParser
from arjuna.interact.gui.dispatcher.selenium.driver import SeleniumDriverDispatcher
from arjuna.interact.gui.auto.finder.emd import GuiElementMetaData

class GuiAutomator(ElementContainer,Dispatchable):

    def __init__(self, app, config, ext_config=None):
        ElementContainer.__init__(self, config)
        Dispatchable.__init__(self)
        self.__app = app
        self.__econfig = ext_config
        self.__create_screenshots_dir()
        self.__main_window = None
        self.__in_slomo = config.get_arjuna_option_value(ArjunaOption.GUIAUTO_SLOMO_ON).as_bool()
        self.__slomo_interval = config.get_arjuna_option_value(ArjunaOption.GUIAUTO_SLOMO_INTERVAL).as_int()

        from .webalert_handler import WebAlertHandler
        from .automator_conditions import GuiAutomatorConditions
        from .viewcontext_handler import ViewContextHandler
        self.__alert_handler = WebAlertHandler(self)
        self.__conditions_handler = GuiAutomatorConditions(self)
        self.__view_handler = ViewContextHandler(self)
        self.__browser = None

        self.__source_parser = None

        # As of now it directly connects to Selenium Dispatcher
        # Code should be introduced here which passes through DispatcherPicker
        # based on choice of engine to support more libs.
        self.dispatcher = SeleniumDriverDispatcher()
        self.__launch()

    @property
    def app(self):
        return self.__app

    @property
    def ext_config(self):
        return self.__econfig

    def create_lmd(self, *locators):
        return GuiElementMetaData.create_lmd(*locators)

    def get_source_from_remote(self):
        return self.dispatcher.get_source()

    def load_source_parser(self):
        raw_source = self.get_source_from_remote()
        self.__source_parser = ElementXMLSourceParser(raw_source, root_element="html")
        self.__source_parser.load()

    # def create_dispatcher(self):
    #     self._set_dispatcher(self.dispatcher_creator.create_gui_automator_dispatcher(self.config, self.setu_id))

    def slomo(self):
        if self.__in_slomo:
            time.sleep(self.__slomo_interval)

    def set_slomo(self, on, interval=None):
        self.__in_slomo = on
        if interval is not None:
            self.__slomo_interval = interval

    @property
    def browser(self):
        return self.__browser

    @property
    def main_window(self):
        return self.__main_window

    def child_window(self, lmd):
        return self.main_window.child_window(lmd)

    @property
    def latest_child_window(self):
        return self.main_window.latest_child_window

    def close_all_child_windows(self):
        self.main_window.close_all_child_windows()

    def dom_root(self, gui):
        from arjuna.interact.gui.auto.component.frame import DomRoot
        return DomRoot(gui)

    def frame(self, gui, lmd, iconfig=None):
        return self.dom_root(gui).frame(lmd, iconfig=iconfig)
        
    @property
    def alert_handler(self):
        return self.__alert_handler

    @property
    def view_handler(self):
        return self.__view_handler

    @property
    def conditions(self):
        return self.__conditions_handler

    def __create_screenshots_dir(self):
        sdir = self.config.get_arjuna_option_value(ArjunaOption.SCREENSHOTS_DIR).as_str()
        if not os.path.isdir(sdir):
            os.makedirs(sdir)

    # #Override
    # def _get_object_uri(self):
    #     return self.__automator_uri

    def __launch(self):
        caps = DriverCapabilities(self.config, self.__econfig)
        self.dispatcher.launch(caps.processed_config)

        from arjuna.interact.gui.auto.component.window import MainWindow
        self.__main_window = MainWindow(self.app, self)

        from .browser import Browser
        self.__browser = Browser(self)

        self.load_source_parser()

    def quit(self):
        self.dispatcher.quit()

    def __screenshot(self):
        switch_view_context = None
        if self.config.value(ArjunaOption.MOBILE_OS_NAME).lower() == "android":
            view_name = self.view_handler.get_current_view_context()   
            if self.view_handler._does_name_represent_web_view(view_name) :
                self.view_handler.switch_to_native_view() 
                switch_view_context = view_name

        response = self.dispatcher.take_screenshot()

        if switch_view_context:
            self.view_handler.switch_to_view_context(switch_view_context)
        
        return response

    def take_screenshot(self):
        response = self.__screenshot()
        image = base64.b64decode(response["data"]["codedImage"])
        path = os.path.join(self.config.value(ArjunaOption.SCREENSHOTS_DIR), "{}.png".format(str(time.time()).replace(".", "-")))
        f = open(path, "wb")
        f.write(image)
        f.close()

    def focus_on_main_window(self):
        self.main_window.focus()

    @property
    def source(self): #, reload=True):
        return self.__source_parser

    def perform_action_chain(self, single_action_chain):
        from arjuna.interact.gui.auto.automator.actions import SingleActionChain
        action_chain = SingleActionChain(self)
        action_chain.perform(single_action_chain)

    def find_element_with_js(self, js):
        return self.dispatcher.find_element_with_js(js)

    def find_multielement_with_js(self, js):
        return self.dispatcher.find_multielement_with_js(js)


    '''
        Public API
    '''

    #### Element Finding

    def element(self, gui, lmd, iconfig=None, max_wait_time=None):
        from arjuna.interact.gui.auto.element.guielement import GuiElement
        gui_element = GuiElement(gui, lmd, iconfig=iconfig) 
        self.load_element(gui_element, max_wait_time=max_wait_time)
        return gui_element

    element_with_lmd = element

    def multi_element(self, gui, lmd, iconfig=None, max_wait_time=None):
        from arjuna.interact.gui.auto.element.multielement import GuiMultiElement
        m_guielement = GuiMultiElement(gui, lmd, iconfig=iconfig)
        self.load_multielement(m_guielement, max_wait_time=max_wait_time)
        return m_guielement

    multi_element_with_lmd = multi_element

    def dropdown(self, gui, lmd, option_container_lmd=None, option_lmd=None, iconfig=None):
        from arjuna.interact.gui.auto.component.dropdown import GuiWebSelect
        return GuiWebSelect(gui, lmd, option_container_lmd=option_container_lmd, option_lmd=option_lmd, iconfig=iconfig)

    def radio_group(self, gui, lmd, iconfig=None):
        from arjuna.interact.gui.auto.component.radio_group import GuiWebRadioGroup
        return GuiWebRadioGroup(gui, lmd)

    def tab_group(self, gui, lmd, *, tab_header_lmd, content_relation_attr, content_relation_type, iconfig=None):
        from arjuna.interact.gui.auto.component.tabs import TabGroup
        return TabGroup(
            gui, 
            lmd, 
            tab_header_lmd=tab_header_lmd, 
            content_relation_attr=content_relation_attr, 
            content_relation_type=content_relation_type,
            iconfig=iconfig
        )

    def execute_javascript(self, js):
        return self.browser.execute_javascript(js)

    ################################
    # Components
    ################################

    @property
    def alert(self):
        return self.alert_handler.create_alert()


from enum import Enum, auto

from arjuna.core.enums import GuiInteractionConfigType
from arjuna.interact.gui.auto.finder.emd import ImplWith

class _GuiInteractionConfigBuilder:

    def __init__(self):
        self.__settings = dict()

    def check_type(self, flag):
        self.__settings[GuiInteractionConfigType.CHECK_TYPE] = flag
        return self

    def check_pre_state(self, flag):
        self.__settings[GuiInteractionConfigType.CHECK_PRE_STATE] = flag
        return self

    def check_post_state(self, flag):
        self.__settings[GuiInteractionConfigType.CHECK_POST_STATE] = flag
        return self

    def build(self):
        return GuiInteractionConfig(self.__settings)


class GuiInteractionConfig:
    '''
        Stores configured values for behavior of actions on a given GuiElement.
    '''

    def __init__(self, settings):
        '''
            settings is a dict of dict of GuiInteractionConfigType key/value pairs
        '''
        # 
        self.__settings = settings

    @property
    def settings(self):
        '''
            Returns the dictionary of configured values for actions configuration.
        '''
        return self.__settings

    @staticmethod
    def builder():
        '''
            Returns a builder object to construct object of this class incrementally.
        '''
        return _GuiInteractionConfigBuilder()

class GuiDriverExtendedConfig:

    def __init__(self, capabilities, browser_args, browser_prefs, browser_exts):
        self.__capabilities = capabilities
        self.__browser_args = browser_args
        self.__browser_prefs = browser_prefs
        self.__browser_exts = browser_exts

    @property
    def config(self):
        '''
            Returns all configuration settings as a single dictionary.
        '''
        map = dict()
        map["driverCapabilities"] = self.__capabilities
        map["browserArgs"] = self.__browser_args 
        map["browserPreferences"] = self.__browser_prefs
        map["browserExtensions"] = self.__browser_exts
        return map

class GuiDriverExtendedConfigBuilder:

    def __init__(self):
        self.__capabilities = dict() 
        self.__browser_args = []
        self.__browser_prefs = dict()
        self.__browser_exts = []

    def capability(self, name, value):
        self.__capabilities[name] = value
        return self

    def browser_arg(self, arg):
        self.__browser_args.append(arg)
        return self

    def browser_pref(self, name, value):
        self.__browser_prefs[name] = value
        return self

    def browser_ext(self, path):
        self.__browser_exts.append(path)
        return self

    def build(self):
        return GuiDriverExtendedConfig(self.__capabilities, self.__browser_args, self.__browser_prefs, self.__browser_exts)


class Keyboard:

    class KeyChord:

        def __init__(self):
            self.__parts = []

        def text(self, text):
            self.__parts.append(text)
            return self

        def key(self, key):
            self.__parts.append(key.name)
            return self

        @property
        def parts(self):
            return self.__parts

    @staticmethod
    def chord(self):
        '''
            Returns a new KeyChord object.
        '''
        return KeyChord()

class _Point:

    def __init__(self, x, y):
        self.__x = x
        self.__y = y

    @property
    def location(self):
        return (self.__x, self.__y)

class _Offset(_Point):

    def __init__(self, x, y):
        super().__init__(x,y)

class Screen:

    @staticmethod
    def xy(x, y):
        return _Point(x,y)

    @staticmethod
    def offset(x, y):
        return _Offset(x,y)

class WithType(Enum):
    ID = auto()
    NAME = auto()
    CLASS_NAME = auto()
    LINK_TEXT = auto()
    LINK_PTEXT = auto()
    XPATH = auto()
    CSS_SELECTOR = auto()
    TAG_NAME = auto()
    COMPOUND_CLASS = auto()
    CLASS_NAMES = auto()

    TEXT = auto()
    TITLE = auto()
    PTEXT = auto()
    BTEXT = auto()
    ETEXT = auto()
    ATTR_VALUE = auto()
    ATTR_PVALUE = auto()
    ATTR_BVALUE = auto()
    ATTR_EVALUE = auto()
    ATTR_WORD = auto()
    TYPE = auto()
    VALUE = auto()
    POINT = auto()
    JAVASCRIPT = auto()

    INDEX = auto()
    WINDOW_TITLE = auto()
    WINDOW_PTITLE = auto()
    CONTENT_LOCATOR = auto()
    GNS_NAME = auto()

class With:

    def __init__(self, with_type, with_value):
        '''

        :param with_type: an enum constant of type WithType
        :param with_value: a string or int or a With object.
        '''
        self.__with_type = with_type
        self.__with_value = None
        self.__has_content_locator = False
        self.__content_locator = None
        self.__format_called = False
        self.__named_args = None

        if with_type == WithType.CONTENT_LOCATOR:
            if not isinstance(with_value, With):
                raise Exception("For identification with element locator, the argument must be a With object.")
            else:
                self.__has_content_locator = True
                self.__content_locator = with_value
        else:
            self.__with_value = with_value

    @property
    def wtype(self):
        return self.__with_type

    @property
    def named_args(self):
        return self.__named_args

    def raise_format_exc(self):
        if (self.__format_called):
            raise Exception("You can not format a With object more than once.")

    def format(self, **kwargs):
        self.raise_format_exc()
        self.__format_called = True
        self.__named_args = kwargs
        return self

    def as_impl_locator(self):

        impl_with = ImplWith(
            wtype = self.__with_type.name,
            wvalue = self.__has_content_locator and self.__content_locator.as_impl_locator() or self.__with_value,
            named_args = self.__named_args,
            has_content_locator = self.__has_content_locator
        )

        return impl_with

    def as_map(self):
        map = dict()
        map["withType"] = self.__with_type.name
        if not self.__has_content_locator:
            map["withValue"] = self.__with_value
        else:
            map["withValue"] = self.__content_locator.as_map()

        if self.__named_args:
            map["named_args"] = self.__named_args
        
        return map

    @staticmethod
    def id(id):
        return With(WithType.ID, id)

    @staticmethod
    def name(name):
        return With(WithType.NAME, name)

    @staticmethod
    def class_name(name):
        return With(WithType.CLASS_NAME, name)

    @staticmethod
    def link_text(text):
        return With(WithType.LINK_TEXT, text)

    @staticmethod
    def link_ptext(text):
        return With(WithType.LINK_PTEXT, text)

    @staticmethod
    def css_selector(selector):
        return With(WithType.CSS_SELECTOR, selector)

    @staticmethod
    def tag_name(selector):
        return With(WithType.TAG_NAME, selector)

    @staticmethod
    def xpath(xpath):
        return With(WithType.XPATH, xpath)

    @staticmethod
    def text(text):
        return With(WithType.TEXT, text)

    @staticmethod
    def ptext(text):
        return With(WithType.PTEXT, text)

    @staticmethod
    def btext(text):
        return With(WithType.BTEXT, text)

    @staticmethod
    def etext(text):
        return With(WithType.ETEXT, text)

    @staticmethod
    def title(title):
        return With(WithType.TITLE, title)

    @staticmethod
    def attr_value(value):
        return With(WithType.ATTR_VALUE, value)

    @staticmethod
    def attr_pvalue(value):
        return With(WithType.ATTR_PVALUE, value)

    @staticmethod
    def attr_bvalue(value):
        return With(WithType.ATTR_BVALUE, value)

    @staticmethod
    def attr_evalue(value):
        return With(WithType.ATTR_EVALUE, value)

    @staticmethod
    def attr_word(value):
        return With(WithType.ATTR_WORD, value)

    @staticmethod
    def value(value):
        return With(WithType.VALUE, value)

    @staticmethod
    def type(type):
        return With(WithType.TYPE, type)

    @staticmethod
    def compound_class(classes_str):
        return With(WithType.COMPOUND_CLASS, classes_str)

    @staticmethod
    def class_names(*class_names):
        return With(WithType.CLASS_NAMES, class_names)

    @staticmethod
    def point(point):
        return With(WithType.POINT, point.location)

    @staticmethod
    def javascript(js):
        return With(WithType.JAVASCRIPT, js)

    @staticmethod
    def index(index):
        return With(WithType.INDEX, index)

    @staticmethod
    def window_title(title):
        return With(WithType.WINDOW_TITLE, title)

    @staticmethod
    def window_ptitle(ptitle):
        return With(WithType.WINDOW_PTITLE, ptitle)

    @staticmethod
    def content_locator(with_obj):
        return With(WithType.CONTENT_LOCATOR, with_obj)

    @staticmethod
    def gns_name(name):
        return With(WithType.GNS_NAME, name)


from enum import Enum
from arjuna.configure.impl.container import ConfigContainer
from arjuna.core.enums import *

class _ConfigCreator:

    def __init__(self, test_session, config_map, conf_trace, code_mode=True):
        self.__code_mode = code_mode
        self.__test_session = test_session
        self.__config_container = ConfigContainer()

        self.__config_map = config_map
        if "default_config" in config_map:
            self.__parent_config = config_map["default_config"]
        else:
            self.__parent_config = None

        # For Unitee
        self.__conf_trace = conf_trace

    def parent_config(self, config):
        self.__parent_config = config

    def arjuna_option(self, option, obj):
        if isinstance(obj, Enum):
            obj = obj.name
        self.__config_container.set_arjuna_option(option, obj)
        return self

    def user_option(self, option, obj):
        self.__config_container.set_user_option(option, obj)
        return self

    def options(self, option_map):
        self.__config_container.set_options(option_map)
        return self

    def selenium(self):
        self.__config_container.set_arjuna_option(ArjunaOption.GUIAUTO_AUTOMATOR_NAME, GuiAutomatorName.SELENIUM.name)
        return self

    def appium(self, context):
        self.arjuna_option(ArjunaOption.GUIAUTO_AUTOMATOR_NAME, GuiAutomatorName.APPIUM.name)
        self.arjuna_option(ArjunaOption.GUIAUTO_CONTEXT, context)
        return self

    def chrome(self):
        self.arjuna_option(ArjunaOption.BROWSER_NAME, BrowserName.CHROME.name)
        return self

    def firefox(self):
        self.arjuna_option(ArjunaOption.BROWSER_NAME, BrowserName.FIREFOX.name)
        return self

    def headless_mode(self):
        self.arjuna_option(ArjunaOption.BROWSER_HEADLESS_MODE, True)
        return self

    def guiauto_max_wait_time(self, seconds):
        self.arjuna_option(ArjunaOption.GUIAUTO_MAX_WAIT, seconds)
        return self

    def app(self, path):
        self.arjuna_option(ArjunaOption.MOBILE_APP_FILE_PATH, path)
        return self

    def mobile_device_name(self, name):
        self.arjuna_option(ArjunaOption.MOBILE_DEVICE_NAME, name)
        return self

    def mobile_device_udid(self, udid):
        self.arjuna_option(ArjunaOption.MOBILE_DEVICE_UDID, udid)
        return self

    def register(self, config_name="default_config"):
        if not self.__config_container.arjuna_options.str_items() and not self.__config_container.user_options.items():
            if not self.__parent_config:
                if config_name != "default_config":
                    self.__config_map[config_name] = self.__config_map["default_config"]
            else:
                self.__config_map[config_name] = self.__parent_config
            return

        config = self.__test_session.register_config(config_name, 
                                        self.__config_container.arjuna_options.str_items(),
                                        self.__config_container.user_options.items(),
                                        self.__parent_config
                                    )

        self.__config_map[config_name] = config
        if self.__code_mode:
            if config_name not in self.__conf_trace:
                self.__conf_trace[config_name] = {"arjuna_options": set(), "user_options" : set()}
            self.__conf_trace[config_name]["arjuna_options"].update(self.__config_container.arjuna_options.items().keys())
            self.__conf_trace[config_name]["user_options"].update(self.__config_container.user_options.items().keys())


class DefaultTestContext:

    def __init__(self, test_session, name, parent_config=None):
        self.__test_session = test_session
        self.__name = name
        self.__parent_config = parent_config and parent_config or None
        from arjuna import Arjuna
        self.__configs = {"default_config" : Arjuna.get_ref_config()}
        self.__conf_trace = dict()

    @property
    def config_creator(self):
        return _ConfigCreator(self.__test_session, self.__configs, self.__conf_trace) # Sent code_mode=True earlier. Check.

    def update_with_file_config_container(self, container):
        for config_name, conf in self.__configs.items():
            builder = self.ConfigBuilder(code_mode=False)
            builder.parent_config(conf)
            amap = container.arjuna_options
            umap = container.user_options
            if config_name in self.__conf_trace:
                if "arjuna_options" in self.__conf_trace[config_name]:
                    for k,v in container.arjuna_options.items():
                        if k not in self.__conf_trace[config_name]["arjuna_options"]:
                            builder.arjuna_option(k, v)
                else:
                    for k, v in container.arjuna_options.items():
                        builder.arjuna_option(k, v)
                if "user_options" in self.__conf_trace[config_name]:
                    for k,v in container.user_options.items():
                        if k not in self.__conf_trace[config_name]["user_options"]:
                            builder.user_option(k, v)
                else:
                    for k, v in container.user_options.items():
                        builder.user_option(k, v)
            else:
                for k in amap.keys():
                    builder.arjuna_option(k, amap.object(k))
                for k in umap.keys():
                        builder.user_option(k, umap.object(k))
            builder.build(config_name=config_name)

    @property
    def config(self):
        return self.get_config()

    def get_config(self, config_name="default_config"):
        return self.__configs[config_name]

    def _add_configs(self, configs):
        self.__configs.update(configs)

    def _add_conf_trace(self, conf_trace):
        self.__conf_trace.update(conf_trace)

    def _get_conf_trace(self):
        return self.__conf_trace

    def _get_configs(self):
        return self.__configs

    def _get_test_session(self):
        return self.__test_session

    def get_name(self):
        return self.__name

# class DefaultTestContext:

#     def __init__(self, test_session, name, parent_context=None):
#         self.__test_session = test_session
#         self.__name = name

#         # dict of name and TestConfig
#         if parent_context is not None:
#             self.__config_map = parent_context.clone_config_map()
#         else:
#             self.__config_map = dict()
#             self.__config_builder = self.ConfigBuilder()
#             builder.ref_config(Arjuna.get_ref_config());
#             builder.build(ConfigBuilder.DEFAULT_CONF_NAME)

#     def ConfigBuilder(self):
#         return ConfigBuilder(self.__test_session, self.__config_map)

#     @property
#     def config(self):
#         return self.__config_map[ConfigBuilder.DEFAULT_CONF_NAME]

#     def get_named_config(self, name):
#         if name is None:
#             raise Exception("Config name was passed as None.")
#         else:
#             try:
#                 return self.get_named_config[name.lower()]
#             except:
#                 raise Exception("No context config found with name: " + name)

#     @property
#     def name(self):
#         return self.__name

#     def _update_options(self, option_map):
#         for conf_name, config in option_map.items():
#             builder = self.ConfigBuilder()
#             builder.ref_config(self.__config_map.get(conf_name))
#             builder.options(option_map)
#             builder.build(conf_name)

#     def clone_config_map(self):
#         out_map = dict()
#         out_map.update(self.__config_map)



'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import argparse
import copy
import abc
import sys
import time

from arjuna.core.utils import sys_utils
from arjuna.core.enums import *
from arjuna.core.enums import *
from arjuna.core.reader.textfile import TextResourceReader
from arjuna.core.types import constants
from arjuna.core.adv.py import *
from arjuna.interface.enums import CommandEnum

from .parser import *
from .command import *

class ArjunaCLI:

    def __init__(self, args):
        super().__init__()
        self._args = args
        self.arg_dict = None
        self.main_command = MainCommand()

        # Create parser for primary commands
        subparsers = self.main_command.create_subparsers()

        # Create re-usable parses for command arguments
        # new_project_parser = NewProjectParser()
        project_parser = ProjectParser()
        run_parser = RunParser()
        # session_parser = SessionParser()
        # group_parser = GroupParser()
        # names_parser = NamesParser()

        # Create primary command handlers
        # self.create_project = CreateProject(subparsers, [new_project_parser])
        self.run_project = RunProject(subparsers, [project_parser, run_parser])
        # self.run_session = RunSession(subparsers, [project_parser, run_parser, session_parser])
        # self.run_group = RunGroup(subparsers, [project_parser, run_parser, group_parser])
        # self.run_names = RunNames(subparsers, [project_parser, run_parser, names_parser])

    def init(self):
        time.sleep(0.1)
        self.arg_dict = self.main_command.convert_to_dict(self._args)
        #self.main_command.execute(self.arg_dict)

    def execute(self):
        command = self.arg_dict['command']
        del self.arg_dict['command']

        if not command:
            print("!!!Fatal Error!!! You did not provide any command.")
            print()
            self.main_command.print_help()
            sys.exit(1)

        # Delegation dictionary for primary command description
        desc_cases = {
            # CommandEnum.LAUNCH_SETU: "Launching Setu",
            # CommandEnum.CREATE_PROJECT: "Creating new project",
            CommandEnum.RUN_PROJECT: "Running the project",
            # CommandEnum.RUN_SESSION: "Running the selected session",
            # CommandEnum.RUN_GROUP: "Running the selected group",
            # CommandEnum.RUN_NAMES: "Running the selected names"
        }

        # Hyphens in commands are replaced with underscores for enum conversion
        # So, create-project is internally referred as CREATE_PROJECT
        command_enum = CommandEnum[command.upper().replace("-", "_")]

        print(desc_cases[command_enum] + "...")

        # Delegation dictionary for primary command choices
        # Respective command object's 'execute' method is the handler.
        execute_cases = {
            # CommandEnum.LAUNCH_SETU: (self.launch_setu.execute,),
            # CommandEnum.CREATE_PROJECT: (self.create_project.execute, ),
            CommandEnum.RUN_PROJECT: (self.run_project.execute, ),
            # CommandEnum.RUN_SESSION: (self.run_session.execute, ),
            # CommandEnum.RUN_GROUP: (self.run_group.execute, ),
            # CommandEnum.RUN_NAMES: (self.run_names.execute, )
        }

        # Delegation using Arjuna's Enum based switch-case equivalent
        switch = EnumSwitch(execute_cases, (self.arg_dict,))
        switch(command_enum)
